import dataclasses

import pytest

import yuio.term
import yuio.theme


class TestColors:
    def test_get_color(self):
        t = yuio.theme.Theme()

        assert t.get_color("code") == yuio.term.Color.FORE_MAGENTA
        assert t.get_color("note") == yuio.term.Color.FORE_GREEN
        assert t.get_color("bold") == yuio.term.Color.STYLE_BOLD
        assert t.get_color("b") == yuio.term.Color.STYLE_BOLD
        assert t.get_color("dim") == yuio.term.Color.STYLE_DIM
        assert t.get_color("d") == yuio.term.Color.STYLE_DIM
        assert t.get_color("normal") == yuio.term.Color.FORE_NORMAL
        assert t.get_color("red") == yuio.term.Color.FORE_RED
        assert t.get_color("green") == yuio.term.Color.FORE_GREEN
        assert t.get_color("yellow") == yuio.term.Color.FORE_YELLOW
        assert t.get_color("blue") == yuio.term.Color.FORE_BLUE
        assert t.get_color("magenta") == yuio.term.Color.FORE_MAGENTA
        assert t.get_color("cyan") == yuio.term.Color.FORE_CYAN

    def test_to_color(self):
        t = yuio.theme.Theme()

        assert t.to_color(None) == yuio.term.Color.NONE
        assert t.to_color("blue") == yuio.term.Color.FORE_BLUE
        assert t.to_color(yuio.term.Color.FORE_MAGENTA) == yuio.term.Color.FORE_MAGENTA

    def test_color_paths(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.term.Color.STYLE_BOLD,
                "x/y": yuio.term.Color.FORE_RED,
                "x/z": yuio.term.Color.FORE_GREEN,
                "y/y": yuio.term.Color.FORE_BLUE,
            }

        a = A()

        assert a.get_color("x") == yuio.term.Color.STYLE_BOLD
        assert (
            a.get_color("x/y") == yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_RED
        )
        assert (
            a.get_color("x/z")
            == yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_GREEN
        )
        assert a.get_color("y/y") == yuio.term.Color.FORE_BLUE

    def test_color_redirect(self):
        class A(yuio.theme.Theme):
            colors = {
                "t_r": yuio.term.Color.FORE_RED,
                "t_g": yuio.term.Color.FORE_GREEN,
                "t_b": yuio.term.Color.STYLE_BOLD,
                "t_d": yuio.term.Color.STYLE_DIM,
                "x": "t_b",
                "x/y": "t_r",
                "y": "t_d",
                "y/y": "x/y",
                "z": ["x/y", "y/y", yuio.term.Color.BACK_CYAN],
            }

        a = A()

        assert a.get_color("x") == yuio.term.Color.STYLE_BOLD
        assert (
            a.get_color("x/y") == yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_RED
        )
        assert a.get_color("y") == yuio.term.Color.STYLE_DIM
        assert (
            a.get_color("y/y")
            == yuio.term.Color.STYLE_DIM
            | yuio.term.Color.STYLE_BOLD
            | yuio.term.Color.FORE_RED
        )
        assert (
            a.get_color("z")
            == yuio.term.Color.STYLE_DIM
            | yuio.term.Color.STYLE_BOLD
            | yuio.term.Color.FORE_RED
            | yuio.term.Color.BACK_CYAN
        )

    def test_color_ctx(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.term.Color.STYLE_BOLD,
                "x:y": yuio.term.Color.FORE_RED,
                "x:y/z": yuio.term.Color.STYLE_DIM,
                "x:a": yuio.term.Color.FORE_GREEN,
                "q/q": yuio.term.Color.FORE_BLUE,
            }

        a = A()

        assert a.get_color("x") == yuio.term.Color.STYLE_BOLD
        assert (
            a.get_color("x:y") == yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_RED
        )
        assert (
            a.get_color("x:a")
            == yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_GREEN
        )
        assert (
            a.get_color("x/a/q:y")
            == yuio.term.Color.STYLE_BOLD | yuio.term.Color.FORE_RED
        )
        assert (
            a.get_color("x:y/z")
            == yuio.term.Color.STYLE_BOLD
            | yuio.term.Color.FORE_RED
            | yuio.term.Color.STYLE_DIM
        )
        assert a.get_color("q/q") == yuio.term.Color.FORE_BLUE
        assert a.get_color("q/q:y") == yuio.term.Color.FORE_BLUE


class TestInheritance:
    def test_simple_inheritance(self):
        class A(yuio.theme.Theme):
            colors = {
                "t_a": yuio.term.Color.FORE_RED,
                "t_ab": yuio.term.Color.FORE_GREEN,
                "t_abc": yuio.term.Color.FORE_BLUE,
            }

        class B(yuio.theme.Theme):
            colors = {
                "t_b": yuio.term.Color.FORE_CYAN,
                "t_ab": yuio.term.Color.FORE_MAGENTA,
                "t_abc": yuio.term.Color.FORE_YELLOW,
            }

        class C(A, B):
            colors = {
                "t_c": yuio.term.Color.FORE_BLACK,
                "t_abc": yuio.term.Color.FORE_WHITE,
            }

        c = C()

        assert C.colors["t_a"] == yuio.term.Color.FORE_RED
        assert C.colors["t_b"] == yuio.term.Color.FORE_CYAN
        assert C.colors["t_c"] == yuio.term.Color.FORE_BLACK
        assert C.colors["t_ab"] == yuio.term.Color.FORE_GREEN
        assert C.colors["t_abc"] == yuio.term.Color.FORE_WHITE

        assert c.get_color("t_a") == yuio.term.Color.FORE_RED
        assert c.get_color("t_b") == yuio.term.Color.FORE_CYAN
        assert c.get_color("t_c") == yuio.term.Color.FORE_BLACK
        assert c.get_color("t_ab") == yuio.term.Color.FORE_GREEN
        assert c.get_color("t_abc") == yuio.term.Color.FORE_WHITE

    def test_color_overrides(self):
        class A(yuio.theme.Theme):
            colors = {
                "t_a": yuio.term.Color.FORE_RED,
                "t_b": yuio.term.Color.FORE_GREEN,
                "t_b_2": yuio.term.Color.FORE_MAGENTA,
            }

            def __init__(self):
                super().__init__()

                self._set_color_if_not_overridden("t_a", yuio.term.Color.BACK_RED)
                self._set_color_if_not_overridden("t_b", yuio.term.Color.BACK_GREEN)
                self.set_color("t_b_2", yuio.term.Color.BACK_MAGENTA)

        class B(A):
            colors = {
                "t_b": yuio.term.Color.FORE_CYAN,
                "t_b_2": yuio.term.Color.FORE_CYAN,
            }

        a = A()
        b = B()

        assert A.colors["t_a"] == yuio.term.Color.FORE_RED
        assert A.colors["t_b"] == yuio.term.Color.FORE_GREEN
        assert A.colors["t_b_2"] == yuio.term.Color.FORE_MAGENTA

        assert a.get_color("t_a") == yuio.term.Color.BACK_RED
        assert a.get_color("t_b") == yuio.term.Color.BACK_GREEN
        assert a.get_color("t_b_2") == yuio.term.Color.BACK_MAGENTA

        assert B.colors["t_a"] == yuio.term.Color.FORE_RED
        assert B.colors["t_b"] == yuio.term.Color.FORE_CYAN
        assert B.colors["t_b_2"] == yuio.term.Color.FORE_CYAN

        assert b.get_color("t_a") == yuio.term.Color.BACK_RED
        assert b.get_color("t_b") == yuio.term.Color.FORE_CYAN
        assert b.get_color("t_b_2") == yuio.term.Color.BACK_MAGENTA

    def test_decoration_overrides(self):
        class A(yuio.theme.Theme):
            msg_decorations = {
                "t_a": "a",
                "t_b": "a",
                "t_b_2": "a",
            }

            def __init__(self):
                super().__init__()

                self._set_msg_decoration_if_not_overridden("t_a", "a2")
                self._set_msg_decoration_if_not_overridden("t_b", "a2")
                self.set_msg_decoration("t_b_2", "a2")

        class B(A):
            msg_decorations = {
                "t_b": "b",
                "t_b_2": "b",
            }

        a = A()
        b = B()

        assert A.msg_decorations["t_a"] == "a"
        assert A.msg_decorations["t_b"] == "a"
        assert A.msg_decorations["t_b_2"] == "a"

        assert a.msg_decorations["t_a"] == "a2"
        assert a.msg_decorations["t_b"] == "a2"
        assert a.msg_decorations["t_b_2"] == "a2"

        assert B.msg_decorations["t_a"] == "a"
        assert B.msg_decorations["t_b"] == "b"
        assert B.msg_decorations["t_b_2"] == "b"

        assert b.msg_decorations["t_a"] == "a2"
        assert b.msg_decorations["t_b"] == "b"
        assert b.msg_decorations["t_b_2"] == "a2"


class TestSetAttrs:
    def test_set_color(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.term.Color.STYLE_BOLD,
            }

        a = A()

        assert a.get_color("x") == yuio.term.Color.STYLE_BOLD
        assert a.get_color("x/y") == yuio.term.Color.STYLE_BOLD
        assert a.get_color("z") == yuio.term.Color.NONE

        a.set_color("x", yuio.term.Color.FORE_RED)
        a.set_color("z", yuio.term.Color.FORE_BLUE)

        assert a.get_color("x") == yuio.term.Color.FORE_RED
        assert a.get_color("x/y") == yuio.term.Color.FORE_RED
        assert a.get_color("z") == yuio.term.Color.FORE_BLUE

    def test_set_color_if_not_overridden(self):
        class A(yuio.theme.Theme):
            colors = {
                "x": yuio.term.Color.STYLE_BOLD,
            }

        a = A()

        with pytest.raises(RuntimeError):
            a._set_color_if_not_overridden("x", yuio.term.Color.FORE_RED)

    def test_set_msg_decoration(self):
        class A(yuio.theme.Theme):
            msg_decorations = {
                "x": "a",
            }

        a = A()

        assert a.msg_decorations["x"] == "a"

        a.set_msg_decoration("x", "b")

        assert a.msg_decorations["x"] == "b"

    def test_set_msg_decoration_if_not_overridden(self):
        class A(yuio.theme.Theme):
            msg_decorations = {
                "x": "a",
            }

        a = A()

        with pytest.raises(RuntimeError):
            a._set_msg_decoration_if_not_overridden("x", "b")


terminal_colors_dark = yuio.term.TerminalColors(
    background=yuio.term.ColorValue.from_hex("#00240A"),
    foreground=yuio.term.ColorValue.from_hex("#FFCFCF"),
    black=yuio.term.ColorValue.from_hex("#000000"),
    red=yuio.term.ColorValue.from_hex("#FF0000"),
    green=yuio.term.ColorValue.from_hex("#00FF00"),
    yellow=yuio.term.ColorValue.from_hex("#FFFF00"),
    blue=yuio.term.ColorValue.from_hex("#0000FF"),
    magenta=yuio.term.ColorValue.from_hex("#FF00FF"),
    cyan=yuio.term.ColorValue.from_hex("#00FFFF"),
    white=yuio.term.ColorValue.from_hex("#FFFFFF"),
    lightness=yuio.term.Lightness.DARK,
)

terminal_colors_light = dataclasses.replace(
    terminal_colors_dark,
    background=yuio.term.ColorValue.from_hex("#FFCFCF"),
    foreground=yuio.term.ColorValue.from_hex("#00240A"),
    lightness=yuio.term.Lightness.LIGHT,
)

terminal_colors_unknown = dataclasses.replace(
    terminal_colors_dark,
    lightness=yuio.term.Lightness.UNKNOWN,
)


class TestDefaultTheme:
    def test_no_color_overrides(self):
        term = yuio.term.Term(None, None)  # type: ignore
        theme = yuio.theme.DefaultTheme(term)
        assert (
            theme.get_color("low_priority_color_a") == yuio.term.Color.FORE_NORMAL_DIM
        )
        assert (
            theme.get_color("low_priority_color_b") == yuio.term.Color.FORE_NORMAL_DIM
        )

    def test_no_color_overrides_unknown_lightness(self):
        term = yuio.term.Term(
            None,  # type: ignore
            None,  # type: ignore
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            terminal_colors=terminal_colors_unknown,
        )
        theme = yuio.theme.DefaultTheme(term)
        assert (
            theme.get_color("low_priority_color_a") == yuio.term.Color.FORE_NORMAL_DIM
        )
        assert (
            theme.get_color("low_priority_color_b") == yuio.term.Color.FORE_NORMAL_DIM
        )
        assert theme.get_color(
            "task/progressbar/done/start"
        ) == yuio.term.Color.fore_from_hex("#0000FF")
        assert theme.get_color(
            "task/progressbar/done/end"
        ) == yuio.term.Color.fore_from_hex("#FF00FF")

    def test_dark_theme(self):
        term = yuio.term.Term(
            None,  # type: ignore
            None,  # type: ignore
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            terminal_colors=terminal_colors_dark,
        )
        theme = yuio.theme.DefaultTheme(term)
        assert theme.get_color("low_priority_color_a") == yuio.term.Color.fore_from_hex(
            "#655151"
        )
        assert theme.get_color("low_priority_color_b") == yuio.term.Color.fore_from_hex(
            "#5A4949"
        )
        assert theme.get_color(
            "task/progressbar/done/start"
        ) == yuio.term.Color.fore_from_hex("#0000FF")
        assert theme.get_color(
            "task/progressbar/done/end"
        ) == yuio.term.Color.fore_from_hex("#FF00FF")

    def test_light_theme(self):
        term = yuio.term.Term(
            None,  # type: ignore
            None,  # type: ignore
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            terminal_colors=terminal_colors_light,
        )
        theme = yuio.theme.DefaultTheme(term)
        assert theme.get_color("low_priority_color_a") == yuio.term.Color.fore_from_hex(
            "#00B231"
        )
        assert theme.get_color("low_priority_color_b") == yuio.term.Color.fore_from_hex(
            "#00BF35"
        )
        assert theme.get_color(
            "task/progressbar/done/start"
        ) == yuio.term.Color.fore_from_hex("#0000FF")
        assert theme.get_color(
            "task/progressbar/done/end"
        ) == yuio.term.Color.fore_from_hex("#FF00FF")



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

'''
This module provides base functionality to build CLI interfaces.

Creating and running an app
---------------------------

Yuio's CLI applications have functional interface. Decorate main function
with the :func:`app` decorator, and use :meth:`App.run` method to start it:

.. code-block:: python

    # Let's define an app with one flag and one positional argument.
    @app
    def main(
        #: help message for `arg`.
        arg: str = positional(),
        #: help message for `--flag`.
        flag: int = 0
    ):
        """this command does a thing."""
        yuio.io.info("flag=%r, arg=%r", flag, arg)

    if __name__ == "__main__":
        # We can now use `main.run` to parse arguments and invoke `main`.
        # Notice that `run` does not return anything. Instead, it terminates
        # python process with an appropriate exit code.
        main.run("--flag 10 foobar!".split())

Function's arguments will become program's flags and positionals, and function's
docstring will become app's :attr:`~App.description`.

Help messages for the flags are parsed from line comments
right above the field definition (comments must start with ``#:``).
They are all formatted using Markdown (see :mod:`yuio.md`).

Parsers for CLI argument values are derived from type hints.
Use the `parser` parameter of the :func:`field` function to override them.

Arguments with bool parsers and parsers that support
:meth:`parsing collections <yuio.parse.Parser.supports_parse_many>`
are handled to provide better CLI experience:

.. invisible-code-block: python

    import pathlib

.. code-block:: python

    @app
    def main(
        # Will create flags `--verbose` and `--no-verbose`.
        verbose: bool = True,

        # Will create a flag with `nargs=*`: `--inputs path1 path2 ...`
        inputs: list[pathlib.Path] = [],
    ):
        ...

.. autofunction:: app

.. autoclass:: App

    .. automethod:: run

    .. method:: command(...)

        The original callable what was wrapped by :func:`app`.


Configuring CLI arguments
-------------------------

Names and types of arguments are determined by names and types of the app function's
arguments. You can use the :func:`field` function to override them:

.. autofunction:: field

.. autodata:: yuio.DISABLED

.. autodata:: yuio.MISSING

.. autodata:: yuio.POSITIONAL

.. autodata:: yuio.OMIT

.. autofunction:: inline

.. autofunction:: positional

.. autoclass:: MutuallyExclusiveGroup


Creating argument groups
------------------------

You can use :class:`~yuio.config.Config` as a type of an app function's parameter.
This will make all of config fields into flags as well. By default, Yuio will use
parameter name as a prefix for all fields in the config; you can override it
with :func:`field` or :func:`inline`:

.. code-block:: python

    class KillCmdConfig(yuio.config.Config):
        # Will be loaded from `--signal`.
        signal: int

        # Will be loaded from `-p` or `--pid`.
        pid: int = field(flags=['-p', '--pid'])

    @app
    def main(
        # `kill_cmd.signal` will be loaded from `--kill-cmd-signal`.
        kill_cmd: KillCmdConfig,

        # `copy_cmd_2.signal` will be loaded from `--kill-signal`.
        kill_cmd_2: KillCmdConfig = field(flags='--kill'),

        # `kill_cmd_3.signal` will be loaded from `--signal`.
        kill_cmd_3: KillCmdConfig = field(flags=''),
    ):
        ...

.. note::

    Positional arguments are not allowed in configs,
    only in apps.


App settings
------------

You can override default usage and help messages as well as control some of the app's
help formatting using its arguments:

.. class:: App
    :noindex:

    .. autoattribute:: prog

    .. autoattribute:: usage

    .. autoattribute:: description

    .. autoattribute:: help

    .. autoattribute:: epilog

    .. autoattribute:: allow_abbrev

    .. autoattribute:: subcommand_required

    .. autoattribute:: setup_logging


Creating sub-commands
---------------------

You can create multiple sub-commands for the main function
using the :meth:`App.subcommand` method:

.. code-block:: python

    @app
    def main():
        ...

    @main.subcommand
    def do_stuff():
        ...

There is no limit to how deep you can nest subcommands, but for usability reasons
we suggest not exceeding level of sub-sub-commands (``git stash push``, anyone?)

When user invokes a subcommand, the ``main()`` function is called first,
then subcommand. In the above example, invoking our app with subcommand ``push``
will cause ``main()`` to be called first, then ``push()``.

This behavior is useful when you have some global configuration flags
attached to the ``main()`` command. See the `example app`_ for details.

.. _example app: https://github.com/taminomara/yuio/blob/main/examples/app

.. class:: App
    :noindex:

    .. automethod:: subcommand


Controlling how sub-commands are invoked
----------------------------------------

By default, if a command has sub-commands, the user is required to provide
a sub-command. This behavior can be disabled by setting :attr:`App.subcommand_required`
to :data:`False`.

When this happens, we need to understand whether a subcommand was invoked or not.
To determine this, you can accept a special parameter called ``_command_info``
of type :class:`CommandInfo`. It will contain info about the current function,
including its name and subcommand:

.. code-block:: python

    @app
    def main(_command_info: CommandInfo):
        if _command_info.subcommand is not None:
            # A subcommand was invoked.
            ...

You can call the subcommand on your own by using ``_command_info.subcommand``
as a callable:

.. code-block:: python

    @app
    def main(_command_info: CommandInfo):
        if _command_info.subcommand is not None and ...:
            _command_info.subcommand()  # manually invoking a subcommand

If you wish to disable calling the subcommand, you can return :data:`False`
from the main function:

.. code-block:: python

    @app
    def main(_command_info: CommandInfo):
        ...
        # Subcommand will not be invoked.
        return False

.. autoclass:: CommandInfo
   :members:

'''

from __future__ import annotations

import argparse
import contextlib
import dataclasses
import functools
import inspect
import logging
import os
import re
import sys
import types
from dataclasses import dataclass

import yuio
import yuio.complete
import yuio.config
import yuio.io
import yuio.md
import yuio.parse
import yuio.term
import yuio.theme
from yuio import _typing as _t
from yuio.config import MutuallyExclusiveGroup, field, inline, positional

__all__ = [
    "field",
    "inline",
    "positional",
    "MutuallyExclusiveGroup",
    "AppError",
    "CommandInfo",
    "app",
    "App",
]

C = _t.TypeVar("C", bound=_t.Callable[..., None])
C2 = _t.TypeVar("C2", bound=_t.Callable[..., None])


class AppError(Exception):
    """
    An error that you can throw from an app to finish its execution without printing
    a traceback. If ``msg`` is given, it will be printed using :func:`yuio.io.failure`;
    otherwise, the app quietly exits with code ``1``.

    :param msg:
        error message, will be passed to :func:`yuio.io.failure`.
    :param args:
        formatting arguments for the error message.

    """

    def __init__(self, msg: str | None = None, *args: _t.Any):
        self.msg: str | None = msg
        self.args: tuple[_t.Any, ...] = args


@_t.overload
def app(
    *,
    prog: str | None = None,
    usage: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    version: str | None = None,
) -> _t.Callable[[C], App[C]]: ...


@_t.overload
def app(
    command: C,
    /,
    *,
    prog: str | None = None,
    usage: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    version: str | None = None,
) -> App[C]: ...


def app(
    command: _t.Callable[..., None] | None = None,
    /,
    *,
    prog: str | None = None,
    usage: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    version: str | None = None,
) -> _t.Any:
    """
    Create an application.

    This is a decorator that's supposed to be used on the main method
    of the application. This decorator returns an :class:`App` object.

    :param command:
        the main function of the application.
    :param prog:
        overrides program's name, see :attr:`App.prog`.
    :param usage:
        overrides program's usage description, see :attr:`App.usage`.
    :param description:
        overrides program's description, see :attr:`App.description`.
    :param epilog:
        overrides program's epilog, see :attr:`App.epilog`.
    :param version:
        program's version, will be displayed using the `--version` flag.

    """

    def registrar(command: C, /) -> App[C]:
        return App(
            command,
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            version=version,
        )

    if command is None:
        return registrar
    else:
        return registrar(command)


@dataclass(frozen=True, eq=False, match_args=False)
class CommandInfo:
    """
    Data about the invoked command.

    """

    name: str
    """
    Name of the current command.

    If it was invoked by alias,
    this will contains the primary command name.

    For the main function, the name will be set to ``"__main__"``.

    """

    subcommand: CommandInfo | None
    """
    Subcommand of this command, if one was given.

    """

    # Internal, do not use.
    _config: _t.Any = dataclasses.field(repr=False)
    _executed: bool = dataclasses.field(default=False, repr=False)

    def __call__(self) -> _t.Literal[False]:
        """
        Execute this command.

        """

        if self._executed:
            return False
        object.__setattr__(self, "_executed", True)

        if self._config is not None:
            should_invoke_subcommand = self._config._run(self)
            if should_invoke_subcommand is None:
                should_invoke_subcommand = True
        else:
            should_invoke_subcommand = True

        if should_invoke_subcommand and self.subcommand is not None:
            self.subcommand()

        return False


class App(_t.Generic[C]):
    """
    A class that encapsulates app settings and logic for running it.

    It is better to create instances of this class using the :func:`app` decorator,
    as it provides means to decorate the main function and specify all of the app's
    parameters.

    """

    @dataclass(frozen=True)
    class _SubApp:
        app: App[_t.Any]
        name: str
        aliases: list[str] | None = None
        is_primary: bool = False

    def __init__(
        self,
        command: C,
        /,
        *,
        prog: str | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
        version: str | None = None,
    ):
        self.prog: str | None = prog
        """
        Program or subcommand display name.

        By default, inferred from :data:`sys.argv` and subcommand names.

        See `prog <https://docs.python.org/3/library/argparse.html#prog>`_
        in :mod:`argparse`.

        """

        self.usage: str | None = usage
        """
        Program or subcommand synapsis.

        This string will be colorized according to `bash` syntax,
        and then it will be ``%``-formatted with a single keyword argument `prog`.
        If command supports multiple signatures, each of them should be listed
        on a separate string. For example::

            @app
            def main(): ...

            main.usage = \"\"\"
            %(prog)s [-q] [-f] [-m] [<branch>]
            %(prog)s [-q] [-f] [-m] --detach [<branch>]
            %(prog)s [-q] [-f] [-m] [--detach] <commit>
            ...
            \"\"\"

        By default, generated from CLI flags by argparse.

        See `usage <https://docs.python.org/3/library/argparse.html#usage>`_
        in :mod:`argparse`.

        """

        if not description and command.__doc__:
            description = command.__doc__

        self.description: str | None = description
        """
        Text that is shown before CLI flags help, usually contains
        short description of the program or subcommand.

        The text should be formatted using markdown. For example:

        .. code-block:: python

           @app
           def main(): ...

           main.description = \"\"\"
           this command does a thing.

           # different ways to do a thing

           this command can apply multiple algorithms to achieve
           a necessary state in which a thing can be done. This includes:

           - randomly turning the screen on and off;

           - banging a head on a table;

           - fiddling with your PCs power cord.

           By default, the best algorithm is determined automatically.
           However, you can hint a preferred algorithm via the `--hint-algo` flag.

           \"\"\"

        By default, inferred from command's docstring.

        See `description <https://docs.python.org/3/library/argparse.html#description>`_
        in :mod:`argparse`.

        """

        if help is yuio.DISABLED:
            help = argparse.SUPPRESS
        elif help is None and description:
            lines = description.split("\n\n", 1)
            help = lines[0].rstrip(".")

        self.help: str | None = help
        """
        Short help message that is shown when listing subcommands.

        By default, inferred from command's docstring.

        See `help <https://docs.python.org/3/library/argparse.html#help>`_
        in :mod:`argparse`.

        """

        self.epilog: str | None = epilog
        """
        Text that is shown after the main portion of the help message.

        Text format is identical to the one for :attr:`~App.description`.

        See `epilog <https://docs.python.org/3/library/argparse.html#epilog>`_
        in :mod:`argparse`.

        """

        self.allow_abbrev: bool = False
        """
        Allow abbreviating CLI flags if that doesn't create ambiguity.

        Disabled by default.

        See `allow_abbrev <https://docs.python.org/3/library/argparse.html#allow-abbrev>`_
        in :mod:`argparse`.

        """

        self.subcommand_required: bool = True
        """
        Require the user to provide a subcommand for this command.

        If this command doesn't have any subcommands, this option is ignored.

        Enabled by default.

        """

        self.setup_logging: bool = True
        """
        If :data:`True`, the app will call :func:`logging.basicConfig` during
        its initialization. Disable this if you want to customize
        logging initialization.

        Disabling this option also removes the ``--verbose`` flag form the CLI.

        """

        self.theme: (
            yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
        ) = None
        """
        A custom theme that will be passed to :func:`yuio.io.setup`
        on application startup.

        """

        self.version: str | None = version
        """
        If not :data:`None`, add ``--version`` flag to the CLI.

        """

        self.__sub_apps: dict[str, App._SubApp] = {}

        if callable(command):
            self.__config_type = _command_from_callable(command)
        else:
            raise TypeError(f"expected a function, got {command}")

        functools.update_wrapper(
            self,  # type: ignore
            command,
            assigned=("__module__", "__name__", "__qualname__", "__doc__"),
            updated=(),
        )

        @functools.wraps(command)
        def command(*args, **kwargs):
            if args:
                names = self.__config_type.__annotations__
                if len(args) > len(names):
                    s = "" if len(names) == 1 else "s"
                    raise TypeError(
                        f"expected at most {len(names)} positional argument{s}, got {len(args)}"
                    )
                for arg, name in zip(args, names):
                    if name in kwargs:
                        raise TypeError(f"argument {name} was given twice")
                    kwargs[name] = arg
            return CommandInfo("__raw__", None, self.__config_type(**kwargs), False)()

        self.command: C = command  # type: ignore
        """
        The original callable what was wrapped by :func:`app`.

        """

    @_t.overload
    def subcommand(
        self,
        /,
        *,
        name: str | None = None,
        aliases: list[str] | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
    ) -> _t.Callable[[C2], App[C2]]: ...

    @_t.overload
    def subcommand(
        self,
        cb: C2,
        /,
        *,
        name: str | None = None,
        aliases: list[str] | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
    ) -> App[C2]: ...

    def subcommand(
        self,
        cb: _t.Callable[..., None] | None = None,
        /,
        *,
        name: str | None = None,
        aliases: list[str] | None = None,
        usage: str | None = None,
        help: str | yuio.Disabled | None = None,
        description: str | None = None,
        epilog: str | None = None,
    ) -> _t.Any:
        """
        Register a subcommand for the given app.

        This method can be used as a decorator, similar to the :func:`app` function.

        :param name:
            allows overriding subcommand's name.
        :param aliases:
            allows adding alias names for subcommand.
        :param usage:
            overrides subcommand's usage description, see :attr:`App.usage`.
        :param help:
            overrides subcommand's short help, see :attr:`App.help`.
            pass :data:`yuio.DISABLED` to hide this subcommand in CLI help message.
        :param description:
            overrides subcommand's description, see :attr:`App.description`.
        :param epilog:
            overrides subcommand's epilog, see :attr:`App.epilog`.

        """

        def registrar(cb: C2, /) -> App[C2]:
            app = App(
                cb,
                usage=usage,
                help=help,
                description=description,
                epilog=epilog,
            )

            app.allow_abbrev = self.allow_abbrev

            main_name = name or yuio.to_dash_case(cb.__name__)
            self.__sub_apps[main_name] = App._SubApp(
                app, main_name, aliases, is_primary=True
            )
            if aliases:
                alias_app = App._SubApp(app, main_name)
                self.__sub_apps.update({alias: alias_app for alias in aliases})

            return app

        if cb is None:
            return registrar
        else:
            return registrar(cb)

    def run(self, args: _t.Sequence[str] | None = None) -> _t.NoReturn:
        """
        Parse arguments, set up :mod:`yuio.io` and :mod:`logging`,
        and run the application.

        :param args:
            command line arguments. If none are given,
            use arguments from :data:`sys.argv`.

        """

        if args is None:
            args = sys.argv[1:]

        if "--yuio-custom-completer--" in args:
            index = args.index("--yuio-custom-completer--")
            yuio.complete._run_custom_completer(
                self.__get_completions(), args[index + 1], args[index + 2]
            )
            sys.exit(0)

        yuio.io.setup(theme=self.theme, wrap_stdio=True)

        parser, subparsers_map = self.__setup_arg_parser()
        namespace = parser.parse_args(args)

        if self.setup_logging:
            logging_level = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}.get(
                namespace.verbosity_level, logging.DEBUG
            )
            logging.basicConfig(handlers=[yuio.io.Handler()], level=logging_level)

        try:
            command = self.__load_from_namespace(namespace)
            command()
            sys.exit(0)
        except AppError as e:
            if e.msg:
                yuio.io.failure(e.msg, *e.args)
            sys.exit(1)
        except (argparse.ArgumentTypeError, argparse.ArgumentError) as e:
            # Make sure we print subcommand's usage, not the main one.
            subcommand_path = self.__get_subcommand_path(namespace)
            subparser = subparsers_map[subcommand_path]
            subparser.error(str(e))
        except Exception as e:
            msg = str(e)
            if "Original traceback:" in msg:
                msg = re.sub(
                    r"\n*?^Original traceback:.*",
                    "",
                    msg,
                    flags=re.MULTILINE | re.DOTALL,
                )
            yuio.io.failure_with_tb("Error: %s", msg)
            sys.exit(3)
        finally:
            yuio.io.restore_streams()

    def __load_from_namespace(self, namespace: argparse.Namespace) -> CommandInfo:
        return self.__load_from_namespace_impl(namespace, "app")

    def __load_from_namespace_impl(
        self, namespace: argparse.Namespace, ns_prefix: str
    ) -> CommandInfo:
        config = self.__config_type._load_from_namespace(namespace, ns_prefix=ns_prefix)
        subcommand = None

        if name := getattr(namespace, ns_prefix + "@subcommand", None):
            sub_app = self.__sub_apps[name]
            subcommand = dataclasses.replace(
                sub_app.app.__load_from_namespace_impl(
                    namespace, f"{ns_prefix}/{sub_app.name}"
                ),
                name=sub_app.name,
            )

        return CommandInfo("__main__", subcommand, _config=config)

    def __get_subcommand_path(self, namespace: argparse.Namespace) -> str:
        return self.__get_subcommand_path_impl(namespace, "app")

    def __get_subcommand_path_impl(self, namespace: argparse.Namespace, ns_prefix: str):
        if name := getattr(namespace, ns_prefix + "@subcommand", None):
            sub_app = self.__sub_apps[name]
            return sub_app.app.__get_subcommand_path_impl(
                namespace, f"{ns_prefix}/{sub_app.name}"
            )
        else:
            return ns_prefix

    def __setup_arg_parser(
        self, parser: argparse.ArgumentParser | None = None
    ) -> tuple[argparse.ArgumentParser, dict[str, argparse.ArgumentParser]]:
        prog = self.prog
        if not prog:
            prog = os.path.basename(sys.argv[0])

        parser = parser or _ArgumentParser(
            prog=self.prog,
            usage=self.usage,
            description=self.description,
            epilog=self.epilog,
            allow_abbrev=self.allow_abbrev,
            add_help=False,
            formatter_class=_HelpFormatter,  # type: ignore
        )

        subparsers_map = {}

        self.__setup_arg_parser_impl(self, parser, "app", prog, subparsers_map)

        return parser, subparsers_map

    def __setup_arg_parser_impl(
        self,
        main_app: App[_t.Any],
        parser: argparse.ArgumentParser,
        ns_prefix: str,
        prog: str,
        subparsers_map: dict[str, argparse.ArgumentParser],
    ):
        subparsers_map[ns_prefix] = parser

        self.__config_type._setup_arg_parser(parser, ns_prefix=ns_prefix)

        if self.__sub_apps:
            subparsers = parser.add_subparsers(
                required=self.subcommand_required,
                dest=ns_prefix + "@subcommand",
                metavar="<subcommand>",
                parser_class=_ArgumentParser,
            )

            for name, sub_app in self.__sub_apps.items():
                if not sub_app.is_primary:
                    continue

                sub_prog = f"{prog} {name}"

                subparser = subparsers.add_parser(
                    name,
                    aliases=sub_app.aliases or [],
                    prog=sub_prog,
                    help=sub_app.app.help,
                    usage=sub_app.app.usage,
                    description=sub_app.app.description,
                    epilog=sub_app.app.epilog,
                    allow_abbrev=self.allow_abbrev,
                    add_help=False,
                    formatter_class=_HelpFormatter,  # type: ignore
                )

                sub_app.app.__setup_arg_parser_impl(
                    main_app,
                    subparser,
                    ns_prefix=f"{ns_prefix}/{name}",
                    prog=sub_prog,
                    subparsers_map=subparsers_map,
                )

        if main_app.__config_type is not self.__config_type:
            main_app.__config_type._setup_arg_parser(
                parser,
                group=parser.add_argument_group(
                    "global options"
                ),  # pyright: ignore[reportArgumentType]
                ns_prefix="app",
            )

        aux = parser.add_argument_group("auxiliary options")
        color = aux.add_mutually_exclusive_group()
        color.add_argument(
            "--force-color",
            help="force-enable colored output",
            action=_NoOpAction,  # Note: `yuio.term` inspects `sys.argv` on its own
            nargs=0,
        )
        color.add_argument(
            "--force-no-color",
            help="force-disable colored output",
            action=_NoOpAction,  # Note: `yuio.term` inspects `sys.argv` on its own
            nargs=0,
        )

        aux.add_argument(
            "-h",
            "--help",
            help="show this help message and exit",
            action=_HelpAction,
            nargs=0,
        )

        if main_app.setup_logging:
            aux.add_argument(
                "-v",
                "--verbose",
                help="increase output verbosity",
                # note the merge function in `_Namespace` for this dest.
                action=_StoreConstAction,
                const=1,
                default=0,
                nargs=0,
                dest="verbosity_level",
            )

        if main_app.version is not None:
            aux.add_argument(
                "-V",
                "--version",
                action=_VersionAction,
                nargs=0,
                version=main_app.version,
                help="show program's version number and exit",
            )

        aux.add_argument(
            "--completions",
            help="generate autocompletion scripts and exit",
            nargs="?",
            action=_make_completions_action(main_app),
        )

    def __get_completions(self) -> yuio.complete._CompleterSerializer:
        serializer = yuio.complete._CompleterSerializer(
            add_help=True, add_version=self.version is not None
        )
        self.__setup_arg_parser(serializer.as_parser())
        return serializer

    def __write_completions(self, shell: str):
        yuio.complete._write_completions(self.__get_completions(), self.prog, shell)


class _NoReprConfig(yuio.config.Config):
    def __repr__(self):
        return "<move along, nothing to see here>"


def _command_from_callable(cb: _t.Callable[..., None]) -> type[yuio.config.Config]:
    sig = inspect.signature(cb)

    dct = {}
    annotations = {}

    accepts_command_info = False

    try:
        docs = yuio._find_docs(cb)
    except Exception:
        yuio._logger.exception(
            "Unable to get documentation for %s",
            cb.__qualname__,
        )
        docs = {}

    for name, param in sig.parameters.items():
        if param.kind not in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
            raise TypeError("positional-only and variadic arguments are not supported")

        if name.startswith("_"):
            if name == "_command_info":
                accepts_command_info = True
                continue
            else:
                raise TypeError(f"unknown special param {name}")

        if param.default is not param.empty:
            field = param.default
        else:
            field = yuio.MISSING
        if not isinstance(field, yuio.config._FieldSettings):
            field = _t.cast(
                yuio.config._FieldSettings, yuio.config.field(default=field)
            )
        if field.default is yuio.MISSING:
            field = dataclasses.replace(field, required=True)
        if name in docs:
            field = dataclasses.replace(field, help=docs[name])

        if param.annotation is param.empty:
            raise TypeError(f"param {name} requires type annotation")

        dct[name] = field
        annotations[name] = param.annotation

    dct["_run"] = _command_from_callable_run_impl(
        cb, list(annotations.keys()), accepts_command_info
    )
    dct["__annotations__"] = annotations
    dct["__module__"] = getattr(cb, "__module__", None)
    dct["__doc__"] = getattr(cb, "__doc__", None)

    return types.new_class(
        cb.__name__,
        (_NoReprConfig,),
        {"_allow_positionals": True},
        exec_body=lambda ns: ns.update(dct),
    )


def _command_from_callable_run_impl(
    cb: _t.Callable[..., None], params: list[str], accepts_command_info
):
    def run(self, command_info):
        kw = {name: getattr(self, name) for name in params}
        if accepts_command_info:
            kw["_command_info"] = command_info
        return cb(**kw)

    return run


class _ArgumentParser(argparse.ArgumentParser):
    def parse_known_args(self, args=None, namespace=None):  # type: ignore
        self._merge_by_dest: dict[str, _t.Callable[[_t.Any, _t.Any], _t.Any]] = {
            action.dest: merge
            for action in self._actions
            if (get_merge := getattr(action, "get_merge", None))
            and (merge := get_merge())
        }
        self._merge_by_dest["verbosity_level"] = lambda l, r: l + r
        if namespace is None:
            namespace = _Namespace(self)
        return super().parse_known_args(args=args, namespace=namespace)

    def error(self, message: str) -> _t.NoReturn:
        self.print_usage()
        yuio.io.failure("Error: %s", message)
        sys.exit(2)


class _Namespace(argparse.Namespace):
    # Since we add flags from main function to all of the subparsers,
    # we need to merge them properly. Otherwise, values from subcommands
    # will override values from the main command: `app --foo=x subcommand --foo=y`
    # will result in `--foo` being just `y`, not `merge(x, y)`. In fact, argparse
    # will override every absent flag with its default: `app --foo x subcommand`
    # will result in `--foo` having a default value.
    def __init__(self, parser: _ArgumentParser):
        self.__parser = parser

    def __setattr__(self, name: str, value: _t.Any):
        if value is yuio.MISSING and hasattr(self, name):
            # Flag was not specified in a subcommand, don't override it.
            return
        if (prev := getattr(self, name, yuio.MISSING)) is not yuio.MISSING and (
            merge := self.__parser._merge_by_dest.get(name)
        ) is not None:
            # Flag was specified in main command and in a subcommand, merge the values.
            value = merge(prev, value)
        return super().__setattr__(name, value)


def _make_completions_action(app: App[_t.Any]):
    class _CompletionsAction(argparse.Action):
        def get_usage(self):
            return False

        parser = yuio.parse.OneOf(
            yuio.parse.Lower(yuio.parse.Str()),
            ["all", "bash", "zsh", "fish", "uninstall"],
        )

        def __init__(self, **kwargs):
            kwargs["metavar"] = self.parser.describe_or_def()
            super().__init__(**kwargs)

        def __call__(self, parser, namespace, value, *args):
            try:
                app._App__write_completions(self.parser.parse(value or "all"))  # type: ignore
            except argparse.ArgumentTypeError as e:
                raise argparse.ArgumentError(self, str(e))
            parser.exit()

    return _CompletionsAction


class _NoOpAction(argparse.Action):
    def get_usage(self):
        return False

    def __call__(self, parser, namespace, value, *args):
        pass


class _StoreConstAction(argparse.Action):
    def get_usage(self):
        return False

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.const)


class _HelpAction(argparse.Action):
    def get_usage(self):
        return False

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        parser.exit()


class _VersionAction(argparse.Action):
    def get_usage(self):
        return False

    def __init__(self, version=None, **kwargs):
        super().__init__(**kwargs)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        print(self.version)
        parser.exit()


_MAX_ARGS_COLUMN_WIDTH = 24


class _CliMdFormatter(yuio.md.MdFormatter):  # type: ignore
    def __init__(
        self,
        theme: yuio.theme.Theme,
        *,
        width: int | None = None,
    ):
        self._heading_indent = contextlib.ExitStack()
        self._args_column_width = _MAX_ARGS_COLUMN_WIDTH

        super().__init__(
            theme,
            width=width,
            allow_headings=True,
        )

        self.width = min(self.width, 80)

    def colorize(
        self,
        s: str,
        /,
        *,
        default_color: yuio.term.Color | str = yuio.term.Color.NONE,
    ):
        return yuio.md.colorize(
            self.theme,
            s,
            default_color=default_color,
            parse_cli_flags_in_backticks=True,
        )

    def _format_Heading(self, node: yuio.md.Heading):
        if node.level == 1:
            self._heading_indent.close()

        decoration = self.theme.msg_decorations.get("heading/section", "")
        with self._indent("msg/decoration:heading/section", decoration):
            self._format_Text(
                node,
                default_color=self.theme.get_color(f"msg/text:heading/section"),
            )

        if node.level == 1:
            self._heading_indent.enter_context(self._indent(None, "  "))
        else:
            self._line(self._first_line_indent)

        self._is_first_line = True

    def _format_Usage(self, node: "_Usage"):
        with self._indent(None, node.prefix):
            self._line(
                node.usage.indent(
                    first_line_indent=self._first_line_indent,
                    continuation_indent=self._continuation_indent,
                )
            )

    def _format_HelpArg(self, node: _HelpArg):
        if node.help is None:
            self._line(self._first_line_indent + node.args)
            return

        if node.args.width + 2 > self._args_column_width:
            self._line(self._first_line_indent + node.indent + node.args)
            indent_ctx = self._indent(None, " " * self._args_column_width)
        else:
            indent_ctx = self._indent(
                None,
                node.indent
                + node.args
                + " " * (self._args_column_width - len(node.indent) - node.args.width),
            )

        with indent_ctx:
            if node.help:
                self._format(node.help)

    def _format_HelpArgGroup(self, node: _HelpArgGroup):
        for item in node.items:
            self._format(item)


@dataclass(**yuio._with_slots())
class _Usage(yuio.md.AstBase):
    prefix: yuio.term.ColorizedString
    usage: yuio.term.ColorizedString


@dataclass(**yuio._with_slots())
class _HelpArg(yuio.md.AstBase):
    indent: str
    args: yuio.term.ColorizedString
    help: yuio.md.AstBase | None


@dataclass(**yuio._with_slots())
class _HelpArgGroup(yuio.md.Container[_HelpArg]):
    pass


class _HelpFormatter:
    def __init__(self, prog: str):
        self._prog = prog
        self._term = yuio.io.get_term()
        self._theme = yuio.io.get_theme()
        self._usage_main_color = self._theme.get_color("msg/text:code/sh-usage")
        self._usage_prog_color = self._usage_main_color | self._theme.get_color(
            "hl/prog:sh-usage"
        )
        self._usage_punct_color = self._usage_main_color | self._theme.get_color(
            "hl/punct:sh-usage"
        )
        self._usage_metavar_color = self._usage_main_color | self._theme.get_color(
            "hl/metavar:sh-usage"
        )
        self._usage_flag_color = self._usage_main_color | self._theme.get_color(
            "hl/flag:sh-usage"
        )
        self._formatter = _CliMdFormatter(self._theme)
        self._nodes: list[yuio.md.AstBase] = []
        self._args_column_width = 0

    def start_section(self, heading: str | None):
        if heading:
            if not heading.endswith(":"):
                heading += ":"
            self._nodes.append(
                yuio.md.Heading(lines=[heading], level=1, start=0, end=0)
            )

    def end_section(self):
        if self._nodes and isinstance(self._nodes[-1], yuio.md.Heading):
            self._nodes.pop()

    def add_text(self, text):
        if text != argparse.SUPPRESS and text:
            self._nodes.append(self._formatter.parse(text))

    def add_usage(
        self, usage, actions: _t.Iterable[argparse.Action], groups, prefix=None
    ):
        if usage == argparse.SUPPRESS:
            return

        if prefix is not None:
            c_prefix = self._formatter.colorize(
                prefix,
                default_color="msg/text:heading/section",
            )
        else:
            c_prefix = yuio.term.ColorizedString(
                [self._theme.get_color("msg/text:heading/section"), "usage: "]
            )

        if usage is not None:
            usage = self._formatter._dedent(usage.strip())
            sh_usage_highlighter = yuio.md.SyntaxHighlighter.get_highlighter("sh-usage")

            c_usage = sh_usage_highlighter.highlight(
                self._theme,
                usage,
            ) % {"prog": self._prog}
        else:
            c_usage = yuio.term.ColorizedString(
                [self._usage_prog_color, str(self._prog)]
            )
            c_usage_elems = yuio.term.ColorizedString()

            optionals: list[argparse.Action | argparse._MutuallyExclusiveGroup] = []
            positionals: list[argparse.Action | argparse._MutuallyExclusiveGroup] = []
            for action in actions:
                if action.option_strings:
                    optionals.append(action)
                else:
                    positionals.append(action)
            for group in groups:
                if len(group._group_actions) <= 1:
                    continue
                for arr in [optionals, positionals]:
                    try:
                        start = arr.index(group._group_actions[0])
                    except (ValueError, IndexError):
                        continue
                    else:
                        end = start + len(group._group_actions)
                        if arr[start:end] == group._group_actions:
                            arr[start:end] = [group]

            has_omitted_usages = False
            sep = False
            for arr in optionals, positionals:
                for elem in arr:
                    if isinstance(elem, argparse.Action):
                        usage_settings = getattr(elem, "get_usage", lambda: True)()
                        if usage_settings is yuio.OMIT:
                            has_omitted_usages = True
                            continue
                        if (
                            not usage_settings
                            or elem.help == argparse.SUPPRESS
                            or elem.metavar == argparse.SUPPRESS
                        ):
                            continue
                        if sep:
                            c_usage_elems += self._usage_main_color
                            c_usage_elems += " "
                        self._format_action_short(elem, c_usage_elems)
                        sep = True
                    elif elem._group_actions:
                        group_actions = []
                        for action in elem._group_actions:
                            usage_settings = getattr(
                                action, "get_usage", lambda: True
                            )()
                            if usage_settings is yuio.OMIT:
                                has_omitted_usages = True
                            elif (
                                usage_settings
                                and action.help != argparse.SUPPRESS
                                and action.metavar != argparse.SUPPRESS
                            ):
                                group_actions.append(action)
                        if not group_actions:
                            continue
                        if sep:
                            c_usage_elems += self._usage_main_color
                            c_usage_elems += " "
                        if len(group_actions) == 1:
                            self._format_action_short(group_actions[0], c_usage_elems)
                            sep = True
                        else:
                            for i, action in enumerate(group_actions):
                                if i == 0:
                                    c_usage_elems += self._usage_punct_color
                                    c_usage_elems += "(" if elem.required else "["
                                self._format_action_short(
                                    action, c_usage_elems, in_group=True
                                )
                                if i + 1 < len(group_actions):
                                    c_usage_elems += self._usage_punct_color
                                    c_usage_elems += "|"
                                else:
                                    c_usage_elems += self._usage_punct_color
                                    c_usage_elems += ")" if elem.required else "]"
                                sep = True

            if has_omitted_usages:
                c_usage_elems_prev = c_usage_elems
                c_usage_elems = yuio.term.ColorizedString(
                    [
                        self._usage_punct_color,
                        "[",
                        self._usage_flag_color,
                        "<options>",
                        self._usage_punct_color,
                        "]",
                    ]
                )
                if c_usage_elems_prev:
                    c_usage_elems += self._usage_main_color
                    c_usage_elems += " "
                    c_usage_elems += c_usage_elems_prev

            if c_usage_elems:
                c_usage += self._usage_main_color
                c_usage += " "
                c_usage += c_usage_elems

        self._nodes.append(
            _Usage(
                start=0,
                end=0,
                prefix=c_prefix,
                usage=c_usage,
            )
        )

    def add_argument(self, action: argparse.Action, indent: str = ""):
        if action.help != argparse.SUPPRESS:
            c_usage = yuio.term.ColorizedString()
            sep = False
            if not action.option_strings:
                self._format_action_metavar(action, 0, c_usage)
            for option_string in action.option_strings:
                if sep:
                    c_usage += self._usage_punct_color
                    c_usage += " "
                c_usage += self._usage_flag_color
                c_usage += option_string
                if action.nargs != 0:
                    c_usage += self._usage_punct_color
                    c_usage += " "
                self._format_action_metavar_expl(action, c_usage)
                sep = True

            if self._nodes and isinstance(self._nodes[-1], _HelpArgGroup):
                group = self._nodes[-1]
            else:
                group = _HelpArgGroup(items=[], start=0, end=0)
                self._nodes.append(group)
            group.items.append(
                _HelpArg(
                    start=0,
                    end=0,
                    indent=indent,
                    args=c_usage,
                    help=self._formatter.parse(action.help) if action.help else None,
                )
            )

            arg_width = len(indent) + c_usage.width + 2
            if arg_width <= _MAX_ARGS_COLUMN_WIDTH:
                self._args_column_width = max(self._args_column_width, arg_width)

            try:
                get_subactions = action._get_subactions  # type: ignore
            except AttributeError:
                pass
            else:
                self.add_arguments(get_subactions(), indent + "  ")

    def add_arguments(self, actions, indent: str = ""):
        for action in actions:
            self.add_argument(action, indent)

    def format_help(self) -> str:
        self._formatter._args_column_width = self._args_column_width
        res = yuio.term.ColorizedString()
        for line in self._formatter.format_node(
            yuio.md.Document(items=self._nodes, start=0, end=0)
        ):
            res += line
            res += "\n"
        res += yuio.term.Color()
        return "".join(res.process_colors(self._term))

    def _format_action_short(
        self,
        action: argparse.Action,
        out: yuio.term.ColorizedString,
        in_group: bool = False,
    ):
        if not in_group and not action.required:
            out += self._usage_punct_color
            out += "["

        if action.option_strings:
            out += self._usage_flag_color
            out += action.option_strings[0]
            if action.nargs != 0:
                out += self._usage_punct_color
                out += " "

        self._format_action_metavar_expl(action, out)

        if not in_group and not action.required:
            out += self._usage_punct_color
            out += "]"

    def _format_action_metavar_expl(
        self, action: argparse.Action, out: yuio.term.ColorizedString
    ):
        nargs = action.nargs if action.nargs is not None else 1

        if nargs == argparse.OPTIONAL:
            out += self._usage_punct_color
            out += "["
            self._format_action_metavar(action, 0, out)
            out += self._usage_punct_color
            out += "]"
        elif nargs == argparse.ZERO_OR_MORE:
            out += self._usage_punct_color
            out += "["
            self._format_action_metavar(action, 0, out)
            out += self._usage_punct_color
            out += " ...]"
        elif nargs == argparse.ONE_OR_MORE:
            self._format_action_metavar(action, 0, out)
            out += self._usage_punct_color
            out += " ["
            self._format_action_metavar(action, 1, out)
            out += self._usage_punct_color
            out += " ...]"
        elif nargs == argparse.REMAINDER:
            out += self._usage_main_color
            out += "..."
        elif nargs == argparse.PARSER:
            self._format_action_metavar(action, 1, out)
            out += self._usage_main_color
            out += " ..."
        elif isinstance(nargs, int):
            sep = False
            for i in range(nargs):
                if sep:
                    out += self._usage_punct_color
                    out += " "
                self._format_action_metavar(action, i, out)
                sep = True

    def _format_action_metavar(
        self, action: argparse.Action, n: int, out: yuio.term.ColorizedString
    ):
        metavar_t = action.metavar
        if not metavar_t and action.option_strings:
            metavar_t = f"<{action.option_strings[0]}>"
        if not metavar_t:
            metavar_t = "<value>"
        if isinstance(metavar_t, tuple):
            metavar = metavar_t[n] if n < len(metavar_t) else metavar_t[-1]
        else:
            metavar = metavar_t

        cur_color = None
        is_punctuation = False
        for part in re.split(r"((?:[{}()[\]\\;!&|]|\s)+)", metavar):
            if is_punctuation and cur_color is not self._usage_punct_color:
                cur_color = self._usage_punct_color
                out += self._usage_punct_color
            elif not is_punctuation and cur_color is not self._usage_metavar_color:
                cur_color = self._usage_metavar_color
                out += self._usage_metavar_color
            out += part
            is_punctuation = not is_punctuation

    def _format_args(self, *_):
        # argparse calls this method sometimes
        # to check if given metavar is valid or not (TODO!)
        pass

    def _set_color(self, *_):
        pass

    def _expand_help(self, *_):
        pass



from __future__ import annotations

import enum
from typing import Annotated

import yuio.config
import yuio.parse


class LogLevel(enum.Enum):
    """
    Available parameters for :attr:`~AppConfig.log_level`.

    """

    ERROR = 3
    """
    Prints messages with severity of *error* and above.

    """

    WARNING = 2
    """
    Prints messages with severity of *warning* and above.

    """

    INFO = 1
    """
    Prints messages with severity of *info* and above.

    """

    DEBUG = 0
    """
    Prints messages with severity of *debug* and above.

    """


class AppConfig(yuio.config.Config):
    """
    Main application config.

    """

    log_level: Annotated[LogLevel, yuio.parse.Enum(by_name=True, to_dash_case=True)] = (
        LogLevel.INFO
    )
    """
    Level used for logging.

    """

    strict: bool = False
    """
    Use strict evaluation metrics.

    """

    executor: ExecutorConfig
    """
    Configuration related to executing the algorithm.

    """


class ExecutorConfig(yuio.config.Config):
    """
    Configuration related to executing the algorithm.

    """

    threads: Annotated[int, yuio.parse.Ge(1)] = 4
    """
    Number of threads to use for executing the algorithm.

    """

    gpu: bool = True
    """
    Enable or disable GPU acceleration.

    """



import re

import pytest

import yuio.md


class TestParser:
    @staticmethod
    def d16(s: str) -> str:
        return "".join(l[16:] + "\n" for l in s.splitlines())

    @staticmethod
    def normalize(s: str) -> str:
        return re.sub(r"\s+", " ", s.strip())

    @pytest.fixture
    def parser(self) -> yuio.md._MdParser:
        return yuio.md._MdParser()

    @pytest.mark.parametrize(
        "md,expected",
        [
            (
                """
                \tfoo\tbar
                """,
                """
                (Code '' 'foo bar')
                """,
            ),
            (
                """
                  \tfoo\tbar
                """,
                """
                (Code '' 'foo bar')
                """,
            ),
            (
                """
                  - foo
                \tbar
                """,
                """
                (List
                  (ListItem None
                    (Paragraph 'foo' 'bar')))
                """,
            ),
            (
                """
                  - foo

                \t\tbar
                """,
                """
                (List
                  (ListItem None
                    (Paragraph 'foo')
                    (Code '' 'bar')))
                """,
            ),
            (
                """
                >\t\tfoo
                """,
                """
                (Quote
                  (Code '' '  foo'))
                """,
            ),
            (
                """
                -\t\tfoo
                """,
                """
                (List
                  (ListItem None
                    (Code '' '  foo')))
                """,
            ),
            (
                """
                    foo
                \tbar
                """,
                """
                (Code '' 'foo' 'bar')
                """,
            ),
            (
                """
                - foo
                   - bar
                \t - baz
                """,
                """
                (List
                  (ListItem None
                    (Paragraph 'foo')
                    (List
                      (ListItem None
                        (Paragraph 'bar')
                        (List
                          (ListItem None
                            (Paragraph 'baz')))))))
                """,
            ),
            (
                """
                    foo
                \tbar
                """,
                """
                (Code '' 'foo' 'bar')
                """,
            ),
            (
                """
                1\\. not a list
                \\* not a list
                \\# not a heading
                """,
                """
                (Paragraph
                  '1\\\\. not a list'
                  '\\\\* not a list'
                  '\\\\# not a heading')
                """,
            ),
            (
                """
                1. Item one.
                1. Item two.
                """,
                """
                (List
                  (ListItem 1
                    (Paragraph 'Item one.'))
                  (ListItem 2
                    (Paragraph 'Item two.')))
                """,
            ),
            (
                """
                2. Item two.
                1. Item three.
                """,
                """
                (List
                  (ListItem 2
                    (Paragraph 'Item two.'))
                  (ListItem 3
                    (Paragraph 'Item three.')))
                """,
            ),
            (
                """
                ---
                ***
                ___
                """,
                """
                (ThematicBreak)
                (ThematicBreak)
                (ThematicBreak)
                """,
            ),
            (
                """
                +++

                ===

                --

                **

                __
                """,
                """
                  (Paragraph '+++')
                  (Paragraph '===')
                  (Paragraph '--')
                  (Paragraph '**')
                  (Paragraph '__')
                """,
            ),
            (
                """
                 ***
                  ***
                   ***
                    ***
                """,
                """
                (ThematicBreak)
                (ThematicBreak)
                (ThematicBreak)
                (Code '' '***')
                """,
            ),
            (
                """
                Foo
                    ***
                """,
                """
                (Paragraph 'Foo' ' ***')
                """,
            ),
            (
                """
                _____
                - - -
                _ _ _
                """,
                """
                (ThematicBreak)
                (ThematicBreak)
                (ThematicBreak)
                """,
            ),
            (
                """
                _ _ _ _ a

                a------

                ---a---
                """,
                """
                (Paragraph '_ _ _ _ a')
                (Paragraph 'a------')
                (Paragraph '---a---')
                """,
            ),
            (
                """
                - foo
                ***
                - bar
                """,
                """
                (List
                  (ListItem None
                    (Paragraph 'foo')))
                (ThematicBreak)
                (List
                  (ListItem None
                    (Paragraph 'bar')))
                """,
            ),
            (
                """
                foo
                ***
                bar
                """,
                """
                (Paragraph 'foo')
                (ThematicBreak)
                (Paragraph 'bar')
                """,
            ),
            (
                """
                foo
                ---
                bar
                """,
                """
                (Heading 2 'foo')
                (Paragraph 'bar')
                """,
            ),
            (
                """
                - foo
                - ---
                - bar
                """,
                """
                (List
                  (ListItem None
                    (Paragraph 'foo'))
                  (ListItem None
                    (ThematicBreak))
                  (ListItem None
                    (Paragraph 'bar')))
                """,
            ),
            (
                """
                # foo
                ## foo
                ### foo
                #### foo
                ##### foo
                ###### foo
                """,
                """
                (Heading 1 'foo')
                (Heading 2 'foo')
                (Heading 3 'foo')
                (Heading 4 'foo')
                (Heading 5 'foo')
                (Heading 6 'foo')
                """,
            ),
            (
                """
                ####### foo

                #5 bolt

                #hashtag
                """,
                """
                (Paragraph '####### foo')
                (Paragraph '#5 bolt')
                (Paragraph '#hashtag')
                """,
            ),
            (
                """
                #                  foo
                """,
                """
                (Heading 1 'foo')
                """,
            ),
            (
                """
                 # foo
                  # foo
                   # foo
                    # foo
                """,
                """
                (Heading 1 'foo')
                (Heading 1 'foo')
                (Heading 1 'foo')
                (Code '' '# foo')
                """,
            ),
            (
                """
                ## foo ##
                ###   bar    ###
                # foo ##################################
                ##### foo ##
                """,
                """
                (Heading 2 'foo')
                (Heading 3 'bar')
                (Heading 1 'foo')
                (Heading 5 'foo')
                """,
            ),
            (
                """
                ### foo ### b
                """,
                """
                (Heading 3 'foo ### b')
                """,
            ),
            (
                """
                ****
                ## foo
                ****
                """,
                """
                (ThematicBreak)
                (Heading 2 'foo')
                (ThematicBreak)
                """,
            ),
            # TODO: this is where I left off:
            #       https://spec.commonmark.org/0.31.2/#example-80
        ],
    )
    def test_ast(self, parser: yuio.md._MdParser, md: str, expected: str):
        md = self.d16(md)
        assert (
            self.normalize(parser.parse(md).dump())
            == f"(Document {self.normalize(expected)})"
        )


# I'm pretty sure MD works as expected, barring some rare edge cases.
# Coverage for yuio.md is low priority rn.



import yuio.app
import yuio.io
from time import sleep

@yuio.app.app
def main():
    emails = ["a@example.com", "b@example.com", "c@example.com"]

    with yuio.io.Task("Sending greetings") as task:
        for i, email in enumerate(emails):
            task.comment(email)
            task.progress(i, len(emails))

            send_greeting()

def send_greeting():
    sleep(3)

if __name__ == "__main__":
    main.run()



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
Yuio's primary format for higher-level io is Markdown (well, a reasonably rich subset
of it).


Formatting markdown
-------------------

.. autoclass:: MdFormatter
   :members:

.. autofunction:: colorize


Highlighting code
-----------------

Yuio supports basic code highlighting; it is just enough to format help messages
for CLI, and color tracebacks when an error occurs.

.. autoclass:: SyntaxHighlighter
   :members:


Markdown AST
------------

.. warning::

   This is an experimental API which can change within a minor release.

.. autoclass:: AstBase
   :members:

.. autoclass:: Text
   :members:

.. autoclass:: Container
   :members:

.. autoclass:: Document
   :members:

.. autoclass:: ThematicBreak
   :members:

.. autoclass:: Heading
   :members:

.. autoclass:: Paragraph
   :members:

.. autoclass:: Quote
   :members:

.. autoclass:: Code
   :members:

.. autoclass:: ListItem
   :members:

.. autoclass:: List
   :members:


"""

from __future__ import annotations

import abc
import contextlib
import dataclasses
import math
import os
import re
import shutil
import string
import textwrap
from dataclasses import dataclass

import yuio.term
import yuio.theme
from yuio import _typing as _t

__all__ = [
    "MdFormatter",
    "AstBase",
    "Text",
    "Container",
    "Document",
    "ThematicBreak",
    "Heading",
    "Paragraph",
    "Quote",
    "Code",
    "ListItem",
    "List",
    "MdFormatter",
    "AstBase",
    "Text",
    "Container",
    "Document",
    "ThematicBreak",
    "Heading",
    "Paragraph",
    "Quote",
    "Code",
    "ListItem",
    "List",
    "SyntaxHighlighter",
    "colorize",
    "strip_color_tags",
]

T = _t.TypeVar("T")
TAst = _t.TypeVar("TAst", bound="AstBase")


@_t.final
class MdFormatter:
    """
    A simple markdown formatter suitable for displaying reach text in the terminal.

    All CommonMark block markup except tables is supported:

    - headings:

      .. code-block:: markdown

         # Heading 1
         ## Heading 2

      Yuio has only two levels of headings. Headings past level two will look the same
      as level two headings (you can adjust theme to change this).

      If `allow_headings` is set to false, headings look like paragraphs.

    - lists, numbered lists, quotes:

      .. code-block:: markdown

         -  List item 1,
         -  list item 2.

         1. Numbered list item 1,
         1. numbered list item 2.

         > Quoted text.

    - fenced code blocks with minimal syntax highlighting
      (see :class:`SyntaxHighlighter`):

      .. code-block:: markdown

         ```python
         for i in range(5, 8):
             print(
                f"Hello, world! "
                "This is {i}th day past the apocalypse."
            )
         ```

      Yuio supports ``python``, ``traceback``, ``bash``, and ``diff`` syntaxes.

    Inline markdown only handles inline code blocks:

    .. code-block:: markdown

       This is `code`. It will be rendered as code.
       Other inline styles, such as _italic_, are not supported!

    However, color tags are supported, so you can highlight text as follows:

    .. code-block:: markdown

       This is <c b>bold text</c>. It will be rendered bold.

    """

    def __init__(
        self,
        theme: yuio.theme.Theme,
        *,
        width: int | None = None,
        allow_headings: bool = True,
    ):
        self.width = width
        self.theme: yuio.theme.Theme = theme
        self.allow_headings: bool = allow_headings

        self._is_first_line: bool
        self._out: list[yuio.term.ColorizedString]
        self._first_line_indent: yuio.term.ColorizedString
        self._continuation_indent: yuio.term.ColorizedString

    @property
    def width(self) -> int:
        """
        Target width for soft-wrapping text.

        """

        return self.__width

    @width.setter
    def width(self, width: int | None):
        if width is None:
            width = shutil.get_terminal_size().columns
        self.__width = max(width, 20)

    def format(self, s: str, /) -> list[yuio.term.ColorizedString]:
        """
        Format a markdown document.

        """

        return self.format_node(self.parse(s))

    def parse(self, s: str, /) -> AstBase:
        """
        Parse a markdown document and return an AST node.

        .. warning::

           This is an experimental API which can change within a minor release.

        """

        return _MdParser(self.allow_headings).parse(self._dedent(s))

    def format_node(self, node: AstBase, /) -> list[yuio.term.ColorizedString]:
        """
        Format a parsed markdown document.

        .. warning::

           This is an experimental API which can change within a minor release.

        """

        self._is_first_line = True
        self._out = []
        self._first_line_indent = yuio.term.ColorizedString()
        self._continuation_indent = yuio.term.ColorizedString()

        self._format(node)

        return self._out

    def colorize(
        self,
        s: str,
        /,
        *,
        default_color: yuio.term.Color | str = yuio.term.Color.NONE,
    ):
        """
        Parse and colorize contents of a paragraph.

        Apply ``default_color`` to the entire paragraph, and process color tags
        and backticks within it.

        """

        return colorize(self.theme, s, default_color=default_color)

    @contextlib.contextmanager
    def _indent(
        self,
        color: yuio.term.Color | str | None,
        s: yuio.term.AnyString,
        /,
        *,
        continue_with_spaces: bool = True,
    ):
        color = self.theme.to_color(color)
        indent = yuio.term.ColorizedString(color) + s

        old_first_line_indent = self._first_line_indent
        old_continuation_indent = self._continuation_indent

        if continue_with_spaces:
            continuation_indent = yuio.term.ColorizedString(" " * indent.width)
        else:
            continuation_indent = indent

        self._first_line_indent = self._first_line_indent + indent
        self._continuation_indent = self._continuation_indent + continuation_indent

        try:
            yield
        finally:
            self._first_line_indent = old_first_line_indent
            self._continuation_indent = old_continuation_indent

    @staticmethod
    def _dedent(s: str, /) -> str:
        first_line, *rest = s.split("\n", 1)
        return first_line + ("\n" + textwrap.dedent(rest[0]) if rest else "")

    def _line(self, line: yuio.term.ColorizedString, /):
        self._out.append(line)

        self._is_first_line = False
        self._first_line_indent = self._continuation_indent

    def _format(self, node: AstBase, /):
        getattr(self, f"_format_{node.__class__.__name__.lstrip('_')}")(node)

    def _format_Text(self, node: Text, /, *, default_color: yuio.term.Color):
        s = self.colorize(
            "\n".join(node.lines).strip(),
            default_color=default_color,
        )

        for line in s.wrap(
            self.width,
            first_line_indent=self._first_line_indent,
            continuation_indent=self._continuation_indent,
            preserve_newlines=False,
        ):
            self._line(line)

    def _format_Container(self, node: Container[TAst], /):
        self._is_first_line = True
        for item in node.items:
            if not self._is_first_line:
                self._line(self._first_line_indent)
            self._format(item)

    def _format_Document(self, node: Document, /):
        self._format_Container(node)

    def _format_ThematicBreak(self, _: ThematicBreak):
        decoration = self.theme.msg_decorations.get("thematic_break", "")
        self._line(self._first_line_indent + decoration)

    def _format_Heading(self, node: Heading, /):
        if not self._is_first_line:
            self._line(self._first_line_indent)

        decoration = self.theme.msg_decorations.get(f"heading/{node.level}", "")
        with self._indent(f"msg/decoration:heading/{node.level}", decoration):
            self._format_Text(
                node,
                default_color=self.theme.get_color(f"msg/text:heading/{node.level}"),
            )

        self._line(self._first_line_indent)
        self._is_first_line = True

    def _format_Paragraph(self, node: Paragraph, /):
        self._format_Text(
            node, default_color=self.theme.get_color("msg/text:paragraph")
        )

    def _format_ListItem(self, node: ListItem, /, *, min_width: int = 0):
        decoration = self.theme.msg_decorations.get("list", "")
        if node.number is not None:
            decoration = f"{node.number:>{min_width}}." + " " * (
                yuio.term.line_width(decoration) - min_width - 1
            )
        with self._indent("msg/decoration:list", decoration):
            self._format_Container(node)

    def _format_Quote(self, node: Quote, /):
        decoration = self.theme.msg_decorations.get("quote", "")
        with self._indent(
            "msg/decoration:quote", decoration, continue_with_spaces=False
        ):
            self._format_Container(node)

    def _format_Code(self, node: Code, /):
        s = SyntaxHighlighter.get_highlighter(node.syntax).highlight(
            self.theme,
            "\n".join(node.lines),
        )

        decoration = self.theme.msg_decorations.get("code", "")
        with self._indent("msg/decoration:code", decoration):
            self._line(
                s.indent(
                    first_line_indent=self._first_line_indent,
                    continuation_indent=self._continuation_indent,
                )
            )

    def _format_List(self, node: List, /):
        max_number = max(item.number or 0 for item in node.items)
        min_width = math.ceil(math.log10(max_number)) if max_number > 0 else 1
        self._is_first_line = True
        for item in node.items:
            if not self._is_first_line:
                self._line(self._first_line_indent)
            self._format_ListItem(item, min_width=min_width)


@dataclass(**yuio._with_slots())
class AstBase(abc.ABC):
    """
    Base class for all AST nodes that represent parsed markdown document.

    """

    start: int
    """
    Start line, 0-based.

    """

    end: int
    """
    End line, 0-based.

    """

    def _dump_params(self) -> str:
        s = self.__class__.__name__.lstrip("_")
        for field in dataclasses.fields(self)[2:]:
            if field.repr:
                s += f" {getattr(self, field.name)!r}"
        return s

    def dump(self, indent: str = "") -> str:
        """
        Dump an AST node into a lisp-like text representation.

        """

        return f"{indent}({self._dump_params()})"


@dataclass(**yuio._with_slots())
class Text(AstBase):
    """
    Base class for all text-based AST nodes, i.e. paragraphs.

    """

    lines: list[str] = dataclasses.field(repr=False)
    """
    Text lines as parsed from the original document.

    """

    def dump(self, indent: str = "") -> str:
        s = f"{indent}({self._dump_params()}"
        indent += "  "
        for line in self.lines:
            s += "\n" + indent
            s += repr(line)
        s += ")"
        return s


@dataclass(**yuio._with_slots())
class Container(AstBase, _t.Generic[TAst]):
    """
    Base class for all container-based AST nodes, i.e. list items or quotes.

    This class works as a list of items. Usually it contains arbitrary AST nodes,
    but it can also be limited to specific kinds of nodes via its generic variable.

    """

    items: list[TAst] = dataclasses.field(repr=False)
    """
    Inner AST nodes in the container.

    """

    def dump(self, indent: str = "") -> str:
        s = f"{indent or ''}({self._dump_params()}"
        indent += "  "
        for items in self.items:
            s += "\n"
            s += items.dump(indent)
        s += ")"
        return s


@dataclass(**yuio._with_slots())
class Document(Container[AstBase]):
    """
    Root node that contains the entire markdown document.

    """


@dataclass(**yuio._with_slots())
class ThematicBreak(AstBase):
    """
    Represents a visual break in text, a.k.a. an asterism.

    """


@dataclass(**yuio._with_slots())
class Heading(Text):
    """
    Represents a heading.

    """

    level: int
    """
    Level of the heading, `1`-based.

    """


@dataclass(**yuio._with_slots())
class Paragraph(Text):
    """
    Represents a regular paragraph.

    """


@dataclass(**yuio._with_slots())
class Quote(Container[AstBase]):
    """
    Represents a quotation block.

    """


@dataclass(**yuio._with_slots())
class Code(Text):
    """
    Represents a highlighted block of code.

    """

    syntax: str
    """
    Syntax indicator as parsed form the original document.

    """


@dataclass(**yuio._with_slots())
class ListItem(Container[AstBase]):
    """
    A possibly numbered element of a list.

    """

    number: int | None
    """
    If present, this is the item's number in a numbered list.

    """


@dataclass(**yuio._with_slots())
class List(Container[ListItem]):
    """
    A collection of list items.

    """


_HEADING_RE = re.compile(
    r"""
    ^
    \s{0,3}                     # - Initial indent.
    (?P<marker>\#{1,6})         # - Heading marker.
    (?P<text>\s.*?)?            # - Heading text. Unless empty, text must be separated
                                #   from the heading marker by a space.
    (?:(?<=\s)\#+)?             # - Optional closing hashes. Must be separated from
                                #   the previous content by a space. We use lookbehind
                                #   here, because if the text is empty, the space
                                #   between heading marker and closing hashes will be
                                #   matched by the `text` group.
    \s*                         # - Closing spaces.
    $
    """,
    re.VERBOSE,
)
_SETEXT_HEADING_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<level>-|=)              # - Heading underline.
    \2*                         # - More heading underline.
    \s*                         # - Closing spaces.
    $
    """,
    re.VERBOSE,
)
_LIST_RE = re.compile(
    r"""
    ^
    (?P<marker>
      \s{0,3}                   # - Initial indent.
      (?P<type>[-*+])           # - List marker.
      (?:
          \s(?:\s{0,3}(?=\S))?  # - One mandatory and up to three optional spaces;
                                #   When there are more than three optional spaces,
                                #   we treat then as a list marker followed
                                #   by a single space, followed by a code block.
        | $))                   # - For cases when a list starts with an empty line.
    (?P<text>.*)                # - Text of the first line in the list.
    $
    """,
    re.VERBOSE,
)
_NUMBERED_LIST_RE = re.compile(
    r"""
    ^
    (?P<marker>
      \s{0,3}                   # - Initial indent.
      (?P<number>\d{1,9})       # - Number.
      (?P<type>[.:)])           # - Numbered list marker.
      (?:
          \s(?:\s{0,3}(?=\S))?  # - One mandatory and up to three optional spaces;
                                #   When there are more than three optional spaces,
                                #   we treat then as a list marker followed
                                #   by a single space, followed by a code block.
        | $))                   # - For cases when a list starts with an empty line.
    (?P<text>.*)                # - Text of the first line in the list.
    $
    """,
    re.VERBOSE,
)
_CODE_BACKTICK_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<fence>```+)             # - Backtick fence.
    (?P<syntax>[^`]*)           # - Syntax, can't contain backtick.
    $
    """,
    re.VERBOSE,
)
_CODE_TILDE_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<fence>~~~+)             # - Backtick fence.
    (?P<syntax>.*)              # - Syntax, can be anything.
    $
    """,
    re.VERBOSE,
)
_CODE_FENCE_END_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial indent.
    (?P<fence>~~~+|```+)        # - Fence.
    \s*                         # - Closing spaces.
    $
    """,
    re.VERBOSE,
)
_CODE_RE = re.compile(
    r"""
    ^
    \s{4}                       # - Initial code indent.
    (?P<text>.*)                # - First code line.
    $
    """,
    re.VERBOSE,
)
_QUOTE_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial quote indent.
    >                           # - Quote marker.
    \s?                         # - Optional space after the marker.
    (?P<text>.*)                # - Text of the first line in the quote.
    $
    """,
    re.VERBOSE,
)
_THEMATIC_BREAK_RE = re.compile(
    r"""
    ^
    (?P<indent>\s{0,3})         # - Initial quote indent.
    ([-*_])\s*(\2\s*){2,}       # - At least three break characters separated by spaces.
    $
    """,
    re.VERBOSE,
)
_LINE_FEED_RE = re.compile(r"\r\n|\r|\n")


class _MdParser:
    @dataclass(**yuio._with_slots())
    class Default:
        pass

    @dataclass(**yuio._with_slots())
    class List:
        item_start: int
        item_end: int
        type: str
        marker_len: int
        list: List
        parser: _MdParser
        number: int | None = None

    @dataclass(**yuio._with_slots())
    class Quote:
        start: int
        end: int
        parser: _MdParser

    @dataclass(**yuio._with_slots())
    class Code:
        start: int
        end: int
        lines: list[str]

    @dataclass(**yuio._with_slots())
    class FencedCode:
        start: int
        end: int
        indent: int
        fence_symbol: str
        fence_length: int
        syntax: str
        lines: list[str]

    @dataclass(**yuio._with_slots())
    class Paragraph:
        start: int
        end: int
        lines: list[str]

    State: _t.TypeAlias = Default | List | Quote | Code | FencedCode | Paragraph

    def __init__(self, allow_headings: bool = True):
        self._allow_headings = allow_headings
        self._nodes: list[AstBase] = []
        self._state: _MdParser.State = self.Default()
        self._cur: int = 0

    def _parser(self) -> _MdParser:
        return _MdParser(self._allow_headings)

    @staticmethod
    def _is_blank(s: str) -> bool:
        return not s or s.isspace()

    def parse(self, s: str) -> AstBase:
        s = s.expandtabs(tabsize=4)
        i = 0
        for i, line in enumerate(_LINE_FEED_RE.split(s)):
            self._handle_line(i, line)
        return Document(items=self._finalize(), start=0, end=i)

    def _handle_line(self, i: int, line: str):
        self._cur = i
        getattr(self, f"_handle_line_{self._state.__class__.__name__}")(line)

    def _handle_lazy_line(self, line: str) -> bool:
        return getattr(self, f"_handle_lazy_line_{self._state.__class__.__name__}")(
            line
        )

    def _flush(self):
        getattr(self, f"_flush_{self._state.__class__.__name__}")()

    def _handle_line_List(self, line: str):
        assert type(self._state) is self.List
        if not line or line[: self._state.marker_len].isspace():
            self._state.item_end = self._cur
            self._state.parser._handle_line(self._cur, line[self._state.marker_len :])
        elif (
            (match := _LIST_RE.match(line)) or (match := _NUMBERED_LIST_RE.match(line))
        ) and match.group("type") == self._state.type:
            item = ListItem(
                items=self._state.parser._finalize(),
                number=self._state.number,
                start=self._state.item_start,
                end=self._state.item_end,
            )
            self._state.list.items.append(item)
            self._state.list.end = item.end
            self._state.item_start = self._state.item_end = self._cur
            self._state.marker_len = len(match.group("marker"))
            self._state.parser._handle_line(self._cur, match.group("text"))
            if self._state.number is not None:
                self._state.number += 1
        elif not self._state.parser._handle_lazy_line(line):
            self._flush_List()
            self._handle_line_Default(line)

    def _handle_lazy_line_List(self, line: str) -> bool:
        assert type(self._state) is self.List
        if self._state.parser._handle_lazy_line(line):
            self._state.item_end = self._cur
            return True
        return False

    def _flush_List(self):
        assert type(self._state) is self.List
        item = ListItem(
            items=self._state.parser._finalize(),
            number=self._state.number,
            start=self._state.item_start,
            end=self._state.item_end,
        )
        self._state.list.items.append(item)
        self._state.list.end = item.end
        self._nodes.append(self._state.list)
        self._state = self.Default()

    def _handle_line_Quote(self, line: str):
        assert type(self._state) is self.Quote
        if match := _QUOTE_RE.match(line):
            self._state.end = self._cur
            self._state.parser._handle_line(self._cur, match.group("text"))
        elif self._is_blank(line) or not self._state.parser._handle_lazy_line(line):
            self._flush_Quote()
            self._handle_line_Default(line)

    def _handle_lazy_line_Quote(self, line: str) -> bool:
        assert type(self._state) is self.Quote
        if self._state.parser._handle_lazy_line(line):
            self._state.end = self._cur
            return True
        else:
            return False

    def _flush_Quote(self):
        assert type(self._state) is self.Quote
        self._nodes.append(
            Quote(
                items=self._state.parser._finalize(),
                start=self._state.start,
                end=self._state.end,
            )
        )
        self._state = self.Default()

    def _handle_line_Code(self, line: str):
        assert type(self._state) == self.Code
        if self._is_blank(line) or line.startswith("    "):
            self._state.end = self._cur
            self._state.lines.append(line[4:])
        else:
            self._flush_Code()
            self._handle_line_Default(line)

    def _handle_lazy_line_Code(self, line: str) -> bool:
        assert type(self._state) == self.Code
        return False  # No lazy continuations for code!

    def _flush_Code(self):
        assert type(self._state) == self.Code
        while self._state.lines and self._is_blank(self._state.lines[-1]):
            self._state.lines.pop()
        self._nodes.append(
            Code(
                lines=self._state.lines,
                syntax="",
                start=self._state.start,
                end=self._state.end,
            )
        )
        self._state = self.Default()

    def _handle_line_FencedCode(self, line: str):
        assert type(self._state) == self.FencedCode
        if (
            (match := _CODE_FENCE_END_RE.match(line))
            and match.group("fence")[0] == self._state.fence_symbol
            and len(match.group("fence")) == self._state.fence_length
        ):
            self._flush_FencedCode()
        else:
            if self._state.indent == 0:
                pass
            elif line[: self._state.indent].isspace():
                line = line[self._state.indent :]
            else:
                line = line.lstrip()
            self._state.end = self._cur
            self._state.lines.append(line)

    def _handle_lazy_line_FencedCode(self, line: str) -> bool:
        assert type(self._state) == self.FencedCode
        return False

    def _flush_FencedCode(self):
        assert type(self._state) == self.FencedCode
        self._nodes.append(
            Code(
                lines=self._state.lines,
                syntax=self._state.syntax,
                start=self._state.start,
                end=self._state.end,
            )
        )
        self._state = self.Default()

    def _handle_line_Paragraph(self, line: str):
        assert type(self._state) == self.Paragraph
        if match := _SETEXT_HEADING_RE.match(line):
            level = 1 if match.group("level") == "=" else 2
            self._nodes.append(
                Heading(
                    lines=self._state.lines,
                    level=level,
                    start=self._state.start,
                    end=self._cur,
                )
            )
            self._state = self.Default()
        elif (
            self._is_blank(line)
            or _THEMATIC_BREAK_RE.match(line)
            or (self._allow_headings and _HEADING_RE.match(line))
            or _CODE_BACKTICK_RE.match(line)
            or _CODE_TILDE_RE.match(line)
            or _LIST_RE.match(line)
            or _NUMBERED_LIST_RE.match(line)
            or _QUOTE_RE.match(line)
        ):
            self._flush_Paragraph()
            self._handle_line_Default(line)
        else:
            self._state.end = self._cur
            self._state.lines.append(line)

    def _handle_lazy_line_Paragraph(self, line: str) -> bool:
        assert type(self._state) == self.Paragraph
        if (
            self._is_blank(line)
            or _THEMATIC_BREAK_RE.match(line)
            or (self._allow_headings and _HEADING_RE.match(line))
            or _CODE_BACKTICK_RE.match(line)
            or _CODE_TILDE_RE.match(line)
            or _LIST_RE.match(line)
            or _NUMBERED_LIST_RE.match(line)
            or _QUOTE_RE.match(line)
        ):
            self._flush_Paragraph()
            return False
        else:
            self._state.end = self._cur
            self._state.lines.append(line)
            return True

    def _flush_Paragraph(self):
        assert type(self._state) == self.Paragraph
        self._nodes.append(
            Paragraph(
                lines=self._state.lines, start=self._state.start, end=self._state.end
            )
        )
        self._state = self.Default()

    def _handle_line_Default(self, line: str):
        assert type(self._state) == self.Default
        if self._is_blank(line):
            pass  # do nothing
        elif _THEMATIC_BREAK_RE.match(line):
            self._nodes.append(ThematicBreak(start=self._cur, end=self._cur))
        elif self._allow_headings and (match := _HEADING_RE.match(line)):
            level = len(match.group("marker"))
            self._nodes.append(
                Heading(
                    lines=[match.group("text").strip()],
                    level=level,
                    start=self._cur,
                    end=self._cur,
                )
            )
        elif (match := _CODE_BACKTICK_RE.match(line)) or (
            match := _CODE_TILDE_RE.match(line)
        ):
            indent = len(match.group("indent"))
            syntax = match.group("syntax").strip()
            fence_symbol = match.group("fence")[0]
            fence_length = len(match.group("fence"))
            self._state = self.FencedCode(
                self._cur, self._cur, indent, fence_symbol, fence_length, syntax, []
            )
        elif match := _CODE_RE.match(line):
            self._state = self.Code(self._cur, self._cur, [match.group("text")])
        elif (match := _LIST_RE.match(line)) or (
            match := _NUMBERED_LIST_RE.match(line)
        ):
            indent = len(match.group("marker"))
            list_type = match.group("type")
            number_str = match.groupdict().get("number", None)
            number = int(number_str) if number_str else None
            self._state = self.List(
                self._cur,
                self._cur,
                list_type,
                indent,
                List(items=[], start=self._cur, end=self._cur),
                self._parser(),
                number,
            )
            self._state.parser._handle_line(self._cur, match.group("text"))
        elif match := _QUOTE_RE.match(line):
            self._state = self.Quote(self._cur, self._cur, self._parser())
            self._state.parser._handle_line(self._cur, match.group("text"))
        else:
            self._state = self.Paragraph(self._cur, self._cur, [line])

    def _handle_lazy_line_Default(self, line: str) -> bool:
        assert type(self._state) == self.Default
        return False

    def _flush_Default(self):
        assert type(self._state) == self.Default

    def _finalize(self) -> list[AstBase]:
        self._flush()
        result = self._nodes
        self._nodes = []
        self._cur = 0
        return result


_SYNTAXES: dict[str, SyntaxHighlighter] = {}
"""
Global syntax registry.

"""


class SyntaxHighlighter(abc.ABC):
    @property
    @abc.abstractmethod
    def syntaxes(self) -> list[str]:
        """
        List of syntax names that should be associated with this highlighter.

        The first name in this list is a canonical syntax name,
        i.e. it will be used to look up colors in a theme.

        """

        return []

    @property
    def syntax(self) -> str:
        """
        The primary syntax name for this highlighter.

        This name is used to look up colors in a theme.

        """

        return self.syntaxes[0] if self.syntaxes else "unknown"

    @classmethod
    def register_highlighter(cls, highlighter: SyntaxHighlighter):
        """
        Register a highlighter in a global registry, and allow looking it up
        via the :meth:`~SyntaxHighlighter.get_highlighter` method.

        """

        for syntax in highlighter.syntaxes:
            _SYNTAXES[syntax.lower().replace("_", "-")] = highlighter

    @classmethod
    def get_highlighter(cls, syntax: str, /) -> SyntaxHighlighter:
        """
        Look up highlighter by a syntax name.

        """

        return _SYNTAXES.get(
            syntax.lower().replace("_", "-"),
            _DummySyntaxHighlighter(),
        )

    @abc.abstractmethod
    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.term.Color | str | None = None,
    ) -> yuio.term.ColorizedString:
        """
        Highlight the given code using the given theme.

        """

    def _get_default_color(
        self,
        theme: yuio.theme.Theme,
        default_color: yuio.term.Color | str | None,
    ) -> yuio.term.Color:
        return theme.to_color(default_color) | theme.get_color(
            f"msg/text:code/{self.syntax}"
        )


class _DummySyntaxHighlighter(SyntaxHighlighter):
    @property
    def syntaxes(self) -> list[str]:
        return ["text", "plain-text"]

    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.term.Color | str | None = None,
    ) -> yuio.term.ColorizedString:
        return yuio.term.ColorizedString(
            [
                self._get_default_color(theme, default_color),
                code,
                yuio.term.Color.NONE,
            ]
        )


SyntaxHighlighter.register_highlighter(_DummySyntaxHighlighter())


class _ReSyntaxHighlighter(SyntaxHighlighter):
    def __init__(
        self,
        syntaxes: list[str],
        pattern: _t.StrRePattern,
        str_esc_pattern: _t.StrRePattern | None = None,
    ):
        self._syntaxes = syntaxes
        self._pattern = pattern
        self._str_esc_pattern = str_esc_pattern

    @property
    def syntaxes(self) -> list[str]:
        return self._syntaxes

    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.term.Color | str | None = None,
    ) -> yuio.term.ColorizedString:
        default_color = self._get_default_color(theme, default_color)

        out = yuio.term.ColorizedString()

        last_pos = 0
        for code_unit in self._pattern.finditer(code):
            if last_pos < code_unit.start():
                out += default_color
                out += code[last_pos : code_unit.start()]
            last_pos = code_unit.end()

            for name, text in sorted(code_unit.groupdict().items()):
                if not text:
                    continue
                name = name.split("__", maxsplit=1)[-1]
                str_color = default_color | theme.get_color(f"hl/str:{self.syntax}")
                esc_color = default_color | theme.get_color(f"hl/str/esc:{self.syntax}")
                if self._str_esc_pattern is not None and name == "str":
                    last_escape_pos = 0
                    for escape_unit in self._str_esc_pattern.finditer(text):
                        if last_escape_pos < escape_unit.start():
                            out += str_color
                            out += text[last_escape_pos : escape_unit.start()]
                        last_escape_pos = escape_unit.end()
                        if escape := code_unit.group(0):
                            out += esc_color
                            out += escape
                    if last_escape_pos < len(text):
                        out += str_color
                        out += code[last_escape_pos:]
                else:
                    out += default_color | theme.get_color(f"hl/{name}:{self.syntax}")
                    out += text

        if last_pos < len(code):
            out += default_color
            out += code[last_pos:]

        out += yuio.term.Color.NONE

        return out


SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["py", "py3", "py-3", "python", "python3", "python-3"],
        re.compile(
            r"""
                (?P<kwd>
                    \b(?:                                   # keyword
                      and|as|assert|async|await|break|class|continue|def|del|elif|else|
                      except|False|finally|for|from|global|if|import|in|is|lambda|None|
                      nonlocal|not|or|pass|raise|return|True|try|while|with|yield
                    )\b)
                | (?P<str>
                    [rfu]*(                                 # string prefix
                        '(?:\\.|[^\\'])*(?:'|\n)            # singly-quoted string
                      | "(?:\\.|[^\\"])*(?:"|\n)            # doubly-quoted string
                      | \"""(\\.|[^\\]|\n)*?\"""            # long singly-quoted string
                      | \'''(\\.|[^\\]|\n)*?\'''))          # long doubly-quoted string
                | (?P<lit>
                      \d+(?:\.\d*(?:e[+-]?\d+)?)?           # int or float
                    | \.\d+(?:e[+-]?\d+)?                   # float that starts with dot
                    | 0x[0-9a-fA-F]+                        # hex
                    | 0b[01]+)                              # bin
                | (?P<punct>[{}()[\]\\;|!&])                # punctuation
                | (?P<comment>\#.*$)                        # comment
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    )
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["sh", "bash"],
        re.compile(
            r"""
                (?P<kwd>
                    \b(?:                                   # keyword
                      if|then|elif|else|fi|time|for|in|until|while|do|done|case|
                      esac|coproc|select|function
                    )\b
                  | \[\[                                    # `test` syntax: if [[ ... ]]
                  | \]\])
                | (?P<a0__punct>(?:^|\|\|?|&&|\$\())        # chaining operator: pipe or logic
                  (?P<a1__>\s*)
                  (?P<a2__prog>([\w.@/-]|\\.)+)             # prog
                | (?P<str>
                    '(?:[.\n]*?)*'                          # singly-quoted string
                  | "(?:\\.|[^\\"])*")                      # doubly-quoted string
                | (?P<punct>
                      [{}()[\]\\;!&|]                       # punctuation
                    | <{1,3}                                # input redirect
                    | [12]?>{1,2}(?:&[12])?)                # output redirect
                | (?P<comment>\#.*$)                        # comment
                | (?P<flag>(?<![\w-])-[a-zA-Z0-9_-]+\b)     # flag
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    ),
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["sh-usage", "bash-usage"],
        re.compile(
            r"""
                (?P<kwd>
                  \b(?:                                     # keyword
                    if|then|elif|else|fi|time|for|in|until|while|do|done|case|
                    esac|coproc|select|function
                  )\b)
                | (?P<prog>%\(prog\)s)                      # prog
                | (?P<metavar>(?<=<)[^>]+(?=>))             # metavar
                | (?P<str>
                      '(?:[.\n]*?)*'                        # singly-quoted string
                    | "(?:\\.|[^\\"])*")                    # doubly-quoted string
                | (?P<comment>\#.*$)                        # comment
                | (?P<flag>(?<![\w-])
                      -[-a-zA-Z0-9_]+\b                     # flag
                    | <options>                             # options
                  )
                | (?P<punct>[{}()[\]\\;!&|])                # punctuation
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    )
)
SyntaxHighlighter.register_highlighter(
    _ReSyntaxHighlighter(
        ["diff"],
        re.compile(
            r"""
                (?P<meta>^(\-\-\-|\+\+\+|\@\@)[^\r\n]*$)
                | (?P<added>^\+[^\r\n]*$)
                | (?P<removed>^\-[^\r\n]*$)
            """,
            re.MULTILINE | re.VERBOSE,
        ),
    ),
)


class _TbHighlighter(SyntaxHighlighter):
    @property
    def syntaxes(self) -> list[str]:
        return [
            "tb",
            "traceback",
            "py-tb",
            "py-traceback",
            "python-tb",
            "python-traceback",
        ]

    class _StackColors:
        def __init__(
            self, theme: yuio.theme.Theme, default_color: yuio.term.Color, tag: str
        ):
            self.file_color = default_color | theme.get_color(f"tb/frame/{tag}/file")
            self.file_path_color = default_color | theme.get_color(
                f"tb/frame/{tag}/file/path"
            )
            self.file_line_color = default_color | theme.get_color(
                f"tb/frame/{tag}/file/line"
            )
            self.file_module_color = default_color | theme.get_color(
                f"tb/frame/{tag}/file/module"
            )
            self.code_color = default_color | theme.get_color(f"tb/frame/{tag}/code")
            self.highlight_color = default_color | theme.get_color(
                f"tb/frame/{tag}/highlight"
            )

    _TB_RE = re.compile(
        r"^(?P<indent>[ |+]*)(Stack|Traceback|Exception Group Traceback) \(most recent call last\):$"
    )
    _TB_MSG_RE = re.compile(r"^(?P<indent>[ |+]*)[A-Za-z_][A-Za-z0-9_]*($|:.*$)")
    _TB_LINE_FILE = re.compile(
        r'^[ |+]*File (?P<file>"[^"]*"), line (?P<line>\d+)(?:, in (?P<loc>.*))?$'
    )
    _TB_LINE_HIGHLIGHT = re.compile(r"^[ |+^~-]*$")
    _SITE_PACKAGES = os.sep + "lib" + os.sep + "site-packages" + os.sep
    _LIB_PYTHON = os.sep + "lib" + os.sep + "python"

    def highlight(
        self,
        theme: yuio.theme.Theme,
        code: str,
        default_color: yuio.term.Color | str | None = None,
    ) -> yuio.term.ColorizedString:
        default_color = self._get_default_color(theme, default_color)

        py_highlighter = SyntaxHighlighter.get_highlighter("python")

        heading_color = default_color | theme.get_color("tb/heading")
        message_color = default_color | theme.get_color("tb/message")

        stack_normal_colors = self._StackColors(theme, default_color, "usr")
        stack_lib_colors = self._StackColors(theme, default_color, "lib")
        stack_colors = stack_normal_colors

        res = yuio.term.ColorizedString()

        PLAIN_TEXT, STACK, MESSAGE = 1, 2, 3
        state = PLAIN_TEXT
        stack_indent = ""
        message_indent = ""

        for line in code.splitlines(keepends=True):
            if state is STACK:
                if line.startswith(stack_indent):
                    # We're still in the stack.
                    if match := self._TB_LINE_FILE.match(line):
                        file, line, loc = match.group("file", "line", "loc")

                        if self._SITE_PACKAGES in file or self._LIB_PYTHON in file:
                            stack_colors = stack_lib_colors
                        else:
                            stack_colors = stack_normal_colors

                        res += yuio.term.Color.NONE
                        res += stack_indent
                        res += stack_colors.file_color
                        res += "File "
                        res += stack_colors.file_path_color
                        res += file
                        res += stack_colors.file_color
                        res += ", line "
                        res += stack_colors.file_line_color
                        res += line
                        res += stack_colors.file_color

                        if loc:
                            res += ", in "
                            res += stack_colors.file_module_color
                            res += loc
                            res += stack_colors.file_color

                        res += "\n"
                    elif match := self._TB_LINE_HIGHLIGHT.match(line):
                        res += yuio.term.Color.NONE
                        res += stack_indent
                        res += stack_colors.highlight_color
                        res += line[len(stack_indent) :]
                    else:
                        res += yuio.term.Color.NONE
                        res += stack_indent
                        res += py_highlighter.highlight(
                            theme,
                            line[len(stack_indent) :],
                            stack_colors.code_color,
                        )
                    continue
                else:
                    # Stack has ended, this line is actually a message.
                    state = MESSAGE

            if state is MESSAGE:
                if line and line != "\n" and line.startswith(message_indent):
                    # We're still in the message.
                    res += yuio.term.Color.NONE
                    res += message_indent
                    res += message_color
                    res += line[len(message_indent) :]
                    continue
                else:
                    # Message has ended, this line is actually a plain text.
                    state = PLAIN_TEXT

            if state is PLAIN_TEXT:
                if match := self._TB_RE.match(line):
                    # Plain text has ended, this is actually a heading.
                    message_indent = match.group("indent").replace("+", "|")
                    stack_indent = message_indent + "  "

                    res += yuio.term.Color.NONE
                    res += message_indent
                    res += heading_color
                    res += line[len(message_indent) :]

                    state = STACK
                    continue
                elif match := self._TB_MSG_RE.match(line):
                    # Plain text has ended, this is an error message (without a traceback).
                    message_indent = match.group("indent").replace("+", "|")
                    stack_indent = message_indent + "  "

                    res += yuio.term.Color.NONE
                    res += message_indent
                    res += message_color
                    res += line[len(message_indent) :]

                    state = MESSAGE
                    continue
                else:
                    # We're still in plain text.
                    res += yuio.term.Color.NONE
                    res += line
                    continue

        res += yuio.term.Color.NONE

        return res


SyntaxHighlighter.register_highlighter(_TbHighlighter())


__TAG_RE = re.compile(
    r"""
          <c (?P<tag_open>[a-z0-9 _/@:-]+)>         # yuio.term.Color tag open.
        | </c>                                      # yuio.term.Color tag close.
        | \\(?P<punct>[%(punct)s])                  # Escape character.
        | (?<!`)(`+)(?!`)(?P<code>.*?)(?<!`)\3(?!`) # Inline code block (backticks).
    """
    % {"punct": re.escape(string.punctuation)},
    re.VERBOSE | re.MULTILINE,
)
__NEG_NUM_RE = re.compile(r"^-(0x[0-9a-fA-F]+|0b[01]+|\d+(e[+-]?\d+)?)$")
__FLAG_RE = re.compile(r"^-[-a-zA-Z0-9_]*$")


def colorize(
    theme: yuio.theme.Theme,
    s: str,
    /,
    *,
    default_color: yuio.term.Color | str = yuio.term.Color.NONE,
    parse_cli_flags_in_backticks: bool = False,
) -> yuio.term.ColorizedString:
    """
    Parse and colorize contents of a paragraph.

    Apply ``default_color`` to the entire paragraph, and process color tags
    and backticks within it.

    """

    default_color = theme.to_color(default_color)

    raw: list[str | yuio.term.Color] = []
    raw.append(default_color)

    stack = [default_color]

    last_pos = 0
    for tag in __TAG_RE.finditer(s):
        raw.append(s[last_pos : tag.start()])
        last_pos = tag.end()

        if name := tag.group("tag_open"):
            color = stack[-1]
            for sub_name in name.split():
                sub_name = sub_name.strip()
                color = color | theme.get_color(sub_name)
            raw.append(color)
            stack.append(color)
        elif code := tag.group("code"):
            code = code.replace("\n", " ")
            if code.startswith(" ") and code.endswith(" ") and not code.isspace():
                code = code[1:-1]
            if (
                parse_cli_flags_in_backticks
                and __FLAG_RE.match(code)
                and not __NEG_NUM_RE.match(code)
            ):
                raw.append(stack[-1] | theme.get_color("hl/flag:sh-usage"))
            else:
                raw.append(stack[-1] | theme.get_color("code"))
            raw.append(yuio.term.NoWrap(code))
            raw.append(stack[-1])
        elif len(stack) > 1:
            stack.pop()
            raw.append(stack[-1])

    raw.append(s[last_pos:])

    raw.append(yuio.term.Color.NONE)

    return yuio.term.ColorizedString(raw)


def strip_color_tags(s: str) -> str:
    """
    Remove all color tags from a string.

    """
    return __TAG_RE.sub("", s)



import os.path
import pathlib

import pytest

import yuio.complete


@pytest.mark.parametrize(
    "strings,expected",
    [
        (["ax", "bx"], ""),
        (["abc", "aby"], "ab"),
        (["foobar", "foobity", "foobaz"], "foob"),
    ],
)
def test_commonprefix(strings, expected):
    assert yuio.complete._commonprefix(strings) == expected


@pytest.mark.parametrize(
    "a,b,expected",
    [
        ("", "", 0),
        ("foo", "FOO", 0),
        ("qux", "qax", 1),
        ("bacon", "beacon", 1),
        ("raeson", "reason", 1),
        ("ass", "add", 2),
        ("r/traaaaannnnnnns", "r/traaaaaaannnnnnnnnns", 5),
    ],
)
def test_corrections(a, b, expected):
    assert yuio.complete._corrections(a, b) == expected


class TestCollector:
    @pytest.mark.parametrize(
        "text,pos,expected_prefix,expected_suffix",
        [
            ("foobar", 0, "", "foobar"),
            ("foobar", 3, "foo", "bar"),
            ("foobar", 6, "foobar", ""),
        ],
    )
    def test_init(self, text, pos, expected_prefix, expected_suffix):
        collector = yuio.complete.CompletionCollector(text, pos)
        assert collector.iprefix == ""
        assert collector.prefix == expected_prefix
        assert collector.suffix == expected_suffix
        assert collector.rsuffix == ""
        assert collector.isuffix == ""

    @pytest.mark.parametrize(
        "iprefix,prefix,suffix,isuffix,completions",
        [
            (
                "",
                "",
                "",
                "",
                {
                    "foo": ("", "foo", ""),
                    "bar": ("", "bar", ""),
                    "baz": ("", "baz", ""),
                },
            ),
            (
                "",
                "ba",
                "",
                "",
                {
                    "foo": None,
                    "bar": ("", "bar", ""),
                    "baz": ("", "baz", ""),
                },
            ),
            (
                "",
                "ba",
                "r",
                "",
                {
                    "foo": None,
                    "bar": ("", "bar", ""),
                    "baz": ("", "baz", ""),
                },
            ),
            (
                "",
                "",
                "foo",
                "",
                {
                    "foo": ("", "foo", ""),
                    "bar": ("", "bar", ""),
                    "baz": ("", "baz", ""),
                },
            ),
            (
                "",
                "meh",
                "",
                "",
                {
                    "foo": None,
                    "bar": None,
                    "baz": None,
                },
            ),
            (
                "mew-",
                "",
                "",
                "-quack",
                {
                    "foo": ("mew-", "foo", "-quack"),
                    "bar": ("mew-", "bar", "-quack"),
                    "baz": ("mew-", "baz", "-quack"),
                },
            ),
            (
                "mew-",
                "ba",
                "r",
                "-quack",
                {
                    "foo": None,
                    "bar": ("mew-", "bar", "-quack"),
                    "baz": ("mew-", "baz", "-quack"),
                },
            ),
        ],
    )
    def test_add(self, iprefix, prefix, suffix, isuffix, completions):
        collector = yuio.complete.CompletionCollector("", 0)
        collector.iprefix = iprefix
        collector.prefix = prefix
        collector.suffix = suffix
        collector.isuffix = isuffix

        for c in completions.keys():
            collector.add(c)

        result = [(c.iprefix, c.completion, c.isuffix) for c in collector.finalize()]
        expected = list(
            sorted([c for c in completions.values() if c], key=lambda x: x[1])
        )
        assert result == expected

    @pytest.mark.parametrize(
        "iprefix,prefix,suffix,isuffix,completions,expected",
        [
            ("", "", "", "", ["bar", "baz"], ("", "ba", "")),
            ("", "b", "", "", ["foo", "bar", "baz"], ("", "ba", "")),
            ("", "b", "ar", "", ["foo", "bar", "baz"], ("", "ba", "")),
            (
                "meow-",
                "b",
                "",
                "-quack",
                ["foo", "bar", "baz"],
                ("meow-", "ba", "-quack"),
            ),
        ],
    )
    def test_common_prefix(
        self, iprefix, prefix, suffix, isuffix, completions, expected
    ):
        collector = yuio.complete.CompletionCollector("", 0)
        collector.iprefix = iprefix
        collector.prefix = prefix
        collector.suffix = suffix
        collector.isuffix = isuffix

        for c in completions:
            collector.add(c)

        result = [(c.iprefix, c.completion, c.isuffix) for c in collector.finalize()]
        assert result == [expected]

    def test_common_prefix_different_isuffix_iprefix(self):
        collector = yuio.complete.CompletionCollector("", 0)

        collector.add("foo")

        with collector.save_state():
            collector.iprefix = "a"
            collector.add("foo")

        with collector.save_state():
            collector.isuffix = "a"
            collector.add("foo")

        with collector.save_state():
            collector.iprefix = "a"
            collector.isuffix = "a"
            collector.add("foo")

        result = [(c.iprefix, c.completion, c.isuffix) for c in collector.finalize()]
        assert result == [
            ("", "foo", ""),
            ("a", "foo", ""),
            ("", "foo", "a"),
            ("a", "foo", "a"),
        ]

    def test_groups(self):
        collector = yuio.complete.CompletionCollector("", 0)

        collector.add("b1")
        collector.add("c1")
        collector.add("a1")

        collector.add_group()

        collector.add("c2")
        collector.add("a2")

        collector.add_group(color_tag="yellow")

        collector.add("b3")
        collector.add("x3")
        collector.add("a3")

        collector.add_group(sorted=False)

        collector.add("b4")
        collector.add("c4")
        collector.add("a4")

        result = [(c.completion, c.group_color_tag) for c in collector.finalize()]
        assert result == [
            ("a1", None),
            ("b1", None),
            ("c1", None),
            ("a2", None),
            ("c2", None),
            ("a3", "yellow"),
            ("b3", "yellow"),
            ("x3", "yellow"),
            ("b4", None),
            ("c4", None),
            ("a4", None),
        ]


class TestCorrectingCollector:
    @pytest.mark.parametrize(
        "text,pos,expected_prefix,expected_suffix",
        [
            ("foobar", 0, "", "foobar"),
            ("foobar", 3, "foo", "bar"),
            ("foobar", 6, "foobar", ""),
        ],
    )
    def test_init(self, text, pos, expected_prefix, expected_suffix):
        collector = yuio.complete._CorrectingCollector(text, pos)
        assert collector.iprefix == ""
        assert collector.prefix == expected_prefix
        assert collector.suffix == expected_suffix
        assert collector.rsuffix == ""
        assert collector.isuffix == ""

    @pytest.mark.parametrize(
        "iprefix,prefix,suffix,isuffix,completions",
        [
            (
                "",
                "abc",
                "",
                "",
                {
                    "abx": ("", "abx", "", None),
                    "foobar": None,
                },
            ),
            (
                "",
                "abc",
                "",
                "",
                {
                    "abx": ("", "abx", "", None),
                    "abxy": ("", "abxy", "", "corrected"),
                    "foobar": None,
                    None: ("", "abc", "", "original"),
                },
            ),
        ],
    )
    def test_add(self, iprefix, prefix, suffix, isuffix, completions):
        collector = yuio.complete._CorrectingCollector("", 0)
        collector.iprefix = iprefix
        collector.prefix = prefix
        collector.suffix = suffix
        collector.isuffix = isuffix

        for c in completions.keys():
            if c:
                collector.add(c)

        result = [
            (c.iprefix, c.completion, c.isuffix, c.comment)
            for c in collector.finalize()
        ]
        expected = [c for c in completions.values() if c]
        assert result == expected


def _get_completions(l: list[yuio.complete.Completion]):
    return [
        (
            c.iprefix,
            c.completion,
            c.isuffix,
            c.rsuffix,
            c.rsymbols,
            c.dprefix,
            c.dsuffix,
            c.comment,
        )
        for c in l
    ]


def test_empty():
    completer = yuio.complete.Empty()

    assert completer.complete("", 0) == []
    assert completer.complete("foo", 1) == []


def test_choice():
    completer = yuio.complete.Choice(
        [
            yuio.complete.Option("foo"),
            yuio.complete.Option("bar"),
            yuio.complete.Option("baz", comment="comment"),
        ]
    )

    assert _get_completions(completer.complete("", 0)) == [
        ("", "bar", "", "", "", "", "", None),
        ("", "baz", "", "", "", "", "", "comment"),
        ("", "foo", "", "", "", "", "", None),
    ]

    assert _get_completions(completer.complete("ba", 2)) == [
        ("", "bar", "", "", "", "", "", None),
        ("", "baz", "", "", "", "", "", "comment"),
    ]


def test_list():
    completer = yuio.complete.List(
        yuio.complete.Choice(
            [
                yuio.complete.Option("foo"),
                yuio.complete.Option("bar"),
                yuio.complete.Option("baz", comment="comment"),
            ]
        )
    )

    assert _get_completions(completer.complete("", 0)) == [
        ("", "bar", "", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("", "baz", "", " ", " \t\n\r\x0b\x0c", "", "", "comment"),
        ("", "foo", "", " ", " \t\n\r\x0b\x0c", "", "", None),
    ]

    assert _get_completions(completer.complete("ba", 2)) == [
        ("", "bar", "", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("", "baz", "", " ", " \t\n\r\x0b\x0c", "", "", "comment"),
    ]

    assert _get_completions(completer.complete("bat", 3)) == [
        ("", "bar", "", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("", "baz", "", " ", " \t\n\r\x0b\x0c", "", "", "comment"),
    ]

    assert _get_completions(completer.complete("qux ba duo", 6)) == [
        ("qux ", "bar", " duo", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("qux ", "baz", " duo", " ", " \t\n\r\x0b\x0c", "", "", "comment"),
    ]

    assert _get_completions(completer.complete("foo  bar", 4)) == [
        ("foo ", "baz", " bar", " ", " \t\n\r\x0b\x0c", "", "", "comment"),
    ]

    completer = yuio.complete.List(
        yuio.complete.Choice(
            [
                yuio.complete.Option("foo"),
                yuio.complete.Option("bar"),
                yuio.complete.Option("baz", comment="comment"),
            ]
        ),
        allow_duplicates=True,
    )

    assert _get_completions(completer.complete("foo  bar", 4)) == [
        ("foo ", "bar", " bar", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("foo ", "baz", " bar", " ", " \t\n\r\x0b\x0c", "", "", "comment"),
        ("foo ", "foo", " bar", " ", " \t\n\r\x0b\x0c", "", "", None),
    ]

    completer = yuio.complete.List(
        yuio.complete.Choice(
            [
                yuio.complete.Option("foo"),
                yuio.complete.Option("bar"),
                yuio.complete.Option("baz", comment="comment"),
            ]
        ),
        delimiter=",",
    )

    assert _get_completions(completer.complete("", 0)) == [
        ("", "bar", "", ",", ",", "", "", None),
        ("", "baz", "", ",", ",", "", "", "comment"),
        ("", "foo", "", ",", ",", "", "", None),
    ]

    assert _get_completions(completer.complete("qux,ba,duo", 6)) == [
        ("qux,", "bar", ",duo", "", "", "", "", None),
        ("qux,", "baz", ",duo", "", "", "", "", "comment"),
    ]

    assert _get_completions(completer.complete("qux,ba", 6)) == [
        ("qux,", "bar", "", ",", ",", "", "", None),
        ("qux,", "baz", "", ",", ",", "", "", "comment"),
    ]


def test_tuple():
    completer = yuio.complete.Tuple(
        yuio.complete.Choice(
            [
                yuio.complete.Option("foo"),
                yuio.complete.Option("bar"),
            ]
        ),
        yuio.complete.Choice(
            [
                yuio.complete.Option("qux"),
                yuio.complete.Option("duo"),
            ]
        ),
    )

    assert _get_completions(completer.complete("", 0)) == [
        ("", "bar", "", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("", "foo", "", " ", " \t\n\r\x0b\x0c", "", "", None),
    ]

    assert _get_completions(completer.complete("  ", 2)) == [
        ("  ", "bar", "", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("  ", "foo", "", " ", " \t\n\r\x0b\x0c", "", "", None),
    ]

    assert _get_completions(completer.complete("x ", 2)) == [
        ("x ", "duo", "", " ", " \t\n\r\x0b\x0c", "", "", None),
        ("x ", "qux", "", " ", " \t\n\r\x0b\x0c", "", "", None),
    ]

    assert _get_completions(completer.complete("x y ", 4)) == []


def _get_file_completions(base: str, l: list[yuio.complete.Completion]):
    return list(
        sorted(
            [
                (
                    c.iprefix.replace(base, "__base__"),
                    c.completion.replace(base, "__base__"),
                    c.isuffix.replace(base, "__base__"),
                    c.rsuffix,
                    c.rsymbols,
                    c.dprefix,
                    c.dsuffix,
                    c.comment,
                )
                for c in l
            ]
        )
    )


def test_file(tmpdir):
    root = pathlib.Path(tmpdir)
    root.joinpath("foo.toml").touch()
    root.joinpath("bar.cfg").touch()
    root.joinpath("baz.cfg").symlink_to(root / "bar.cfg")
    root.joinpath("dir1").mkdir()
    root.joinpath("dir2").mkdir()

    completer = yuio.complete.File()

    base = str(root) + os.path.sep

    assert _get_file_completions(base, completer.complete(base, len(base))) == [
        ("__base__", "bar.cfg", "", "", "", "", "", None),
        ("__base__", "baz.cfg", "", "", "", "", "@", None),
        ("__base__", "dir1" + os.path.sep, "", "", "", "", "", None),
        ("__base__", "dir2" + os.path.sep, "", "", "", "", "", None),
        ("__base__", "foo.toml", "", "", "", "", "", None),
    ]

    assert _get_file_completions(
        base, completer.complete(base + "ba", len(base) + 2)
    ) == [
        ("__base__", "bar.cfg", "", "", "", "", "", None),
        ("__base__", "baz.cfg", "", "", "", "", "@", None),
    ]

    completer = yuio.complete.File(extensions=[".toml"])
    assert _get_file_completions(base, completer.complete(base, len(base))) == [
        ("__base__", "dir1" + os.path.sep, "", "", "", "", "", None),
        ("__base__", "dir2" + os.path.sep, "", "", "", "", "", None),
        ("__base__", "foo.toml", "", "", "", "", "", None),
    ]

    completer = yuio.complete.List(yuio.complete.File(), delimiter=";")
    assert _get_file_completions(base, completer.complete(base, len(base))) == [
        ("__base__", "bar.cfg", "", ";", ";", "", "", None),
        ("__base__", "baz.cfg", "", ";", ";", "", "@", None),
        ("__base__", "dir1" + os.path.sep, "", "", ";", "", "", None),
        ("__base__", "dir2" + os.path.sep, "", "", ";", "", "", None),
        ("__base__", "foo.toml", "", ";", ";", "", "", None),
    ]

    assert _get_file_completions(
        base, completer.complete(base + ";xyz", len(base))
    ) == [
        ("__base__", "bar.cfg", ";xyz", "", "", "", "", None),
        ("__base__", "baz.cfg", ";xyz", "", "", "", "@", None),
        ("__base__", "dir1" + os.path.sep, ";xyz", "", "", "", "", None),
        ("__base__", "dir2" + os.path.sep, ";xyz", "", "", "", "", None),
        ("__base__", "foo.toml", ";xyz", "", "", "", "", None),
    ]

    completer = yuio.complete.Dir()
    assert _get_file_completions(
        base, completer.complete(base + "dir", len(base) + 3)
    ) == [
        ("__base__", "dir1" + os.path.sep, "", "", "", "", "", None),
        ("__base__", "dir2" + os.path.sep, "", "", "", "", "", None),
    ]

    assert _get_file_completions(
        base, completer.complete(base + ".", len(base) + 1)
    ) == [
        ("__base__", ".." + os.path.sep, "", "", "", "", "", None),
        ("__base__", "." + os.path.sep, "", "", "", "", "", None),
    ]

    assert ("", "~" + os.path.sep, "", "", "", "", "", None) in _get_file_completions(
        base, completer.complete("~", 1)
    )



import yuio.app
import yuio.config

class Config(yuio.config.Config):
    use_formal_greeting: bool = yuio.config.field(
        default=False,
        flags=["-f", "--formal"],
        env="FORMAL",
    )

    sender_name: str | None = yuio.config.field(
        default=None,
        flags=["-s", "--sender"],
        env="SENDER",
    )

@yuio.app.app
def main(cli_config: Config = yuio.app.inline()):
    ...

    # `config.sender_name` will be loaded from `GREET_SENDER`.
    config = Config.load_from_env(prefix="GREET")

    ...

    _config = config

if __name__ == "__main__":
    main.run()



import logging
import yuio.app
import yuio.io

logger = logging.getLogger("main")

@yuio.app.app
def main(foo: str = "", bar: str = ""):
    logging.basicConfig(
        level=logging.INFO,
        handlers=[yuio.io.Handler()],
    )

    logger.debug("settings: foo=%r, bar=%r", foo, bar)
    logger.info("Running some logic...")

main.setup_logging = False

if __name__ == "__main__":
    main.run()
